# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from . import emulated_bluetooth_device_pb2 as emulated__bluetooth__device__pb2
from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2


class GattDeviceServiceStub(object):
    """You can provide your own GattDevice by implementing this service
    and registering it with the android emulator.

    The device will appear as a real bluetooth device, and you will
    receive callbacks when the bluetooth system wants to
    read, write or observe a characteristic.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.OnCharacteristicReadRequest = channel.unary_unary(
                '/android.emulation.bluetooth.GattDeviceService/OnCharacteristicReadRequest',
                request_serializer=emulated__bluetooth__device__pb2.CharacteristicValueRequest.SerializeToString,
                response_deserializer=emulated__bluetooth__device__pb2.CharacteristicValueResponse.FromString,
                )
        self.OnCharacteristicWriteRequest = channel.unary_unary(
                '/android.emulation.bluetooth.GattDeviceService/OnCharacteristicWriteRequest',
                request_serializer=emulated__bluetooth__device__pb2.CharacteristicValueRequest.SerializeToString,
                response_deserializer=emulated__bluetooth__device__pb2.CharacteristicValueResponse.FromString,
                )
        self.OnCharacteristicObserveRequest = channel.unary_stream(
                '/android.emulation.bluetooth.GattDeviceService/OnCharacteristicObserveRequest',
                request_serializer=emulated__bluetooth__device__pb2.CharacteristicValueRequest.SerializeToString,
                response_deserializer=emulated__bluetooth__device__pb2.CharacteristicValueResponse.FromString,
                )
        self.OnConnectionStateChange = channel.unary_unary(
                '/android.emulation.bluetooth.GattDeviceService/OnConnectionStateChange',
                request_serializer=emulated__bluetooth__device__pb2.ConnectionStateChange.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                )


class GattDeviceServiceServicer(object):
    """You can provide your own GattDevice by implementing this service
    and registering it with the android emulator.

    The device will appear as a real bluetooth device, and you will
    receive callbacks when the bluetooth system wants to
    read, write or observe a characteristic.
    """

    def OnCharacteristicReadRequest(self, request, context):
        """A remote client has requested to read a local characteristic.

        Return the current observed value.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def OnCharacteristicWriteRequest(self, request, context):
        """A remote client has requested to write to a local characteristic.

        Return the current observed value.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def OnCharacteristicObserveRequest(self, request, context):
        """Listens for notifications from the emulated device, the device should
        write to the stream with a response when a change has occurred.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def OnConnectionStateChange(self, request, context):
        """A remote device has been connected or disconnected.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_GattDeviceServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'OnCharacteristicReadRequest': grpc.unary_unary_rpc_method_handler(
                    servicer.OnCharacteristicReadRequest,
                    request_deserializer=emulated__bluetooth__device__pb2.CharacteristicValueRequest.FromString,
                    response_serializer=emulated__bluetooth__device__pb2.CharacteristicValueResponse.SerializeToString,
            ),
            'OnCharacteristicWriteRequest': grpc.unary_unary_rpc_method_handler(
                    servicer.OnCharacteristicWriteRequest,
                    request_deserializer=emulated__bluetooth__device__pb2.CharacteristicValueRequest.FromString,
                    response_serializer=emulated__bluetooth__device__pb2.CharacteristicValueResponse.SerializeToString,
            ),
            'OnCharacteristicObserveRequest': grpc.unary_stream_rpc_method_handler(
                    servicer.OnCharacteristicObserveRequest,
                    request_deserializer=emulated__bluetooth__device__pb2.CharacteristicValueRequest.FromString,
                    response_serializer=emulated__bluetooth__device__pb2.CharacteristicValueResponse.SerializeToString,
            ),
            'OnConnectionStateChange': grpc.unary_unary_rpc_method_handler(
                    servicer.OnConnectionStateChange,
                    request_deserializer=emulated__bluetooth__device__pb2.ConnectionStateChange.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'android.emulation.bluetooth.GattDeviceService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class GattDeviceService(object):
    """You can provide your own GattDevice by implementing this service
    and registering it with the android emulator.

    The device will appear as a real bluetooth device, and you will
    receive callbacks when the bluetooth system wants to
    read, write or observe a characteristic.
    """

    @staticmethod
    def OnCharacteristicReadRequest(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/android.emulation.bluetooth.GattDeviceService/OnCharacteristicReadRequest',
            emulated__bluetooth__device__pb2.CharacteristicValueRequest.SerializeToString,
            emulated__bluetooth__device__pb2.CharacteristicValueResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def OnCharacteristicWriteRequest(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/android.emulation.bluetooth.GattDeviceService/OnCharacteristicWriteRequest',
            emulated__bluetooth__device__pb2.CharacteristicValueRequest.SerializeToString,
            emulated__bluetooth__device__pb2.CharacteristicValueResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def OnCharacteristicObserveRequest(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/android.emulation.bluetooth.GattDeviceService/OnCharacteristicObserveRequest',
            emulated__bluetooth__device__pb2.CharacteristicValueRequest.SerializeToString,
            emulated__bluetooth__device__pb2.CharacteristicValueResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def OnConnectionStateChange(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/android.emulation.bluetooth.GattDeviceService/OnConnectionStateChange',
            emulated__bluetooth__device__pb2.ConnectionStateChange.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -----------------------------------------------------------------------------
# GATT - Type Adapters
# -----------------------------------------------------------------------------

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------
from __future__ import annotations

import struct
from collections.abc import Callable, Iterable
from typing import Any, Generic, Literal, TypeVar

from bumble import utils
from bumble.core import InvalidOperationError
from bumble.gatt import Characteristic
from bumble.gatt_client import CharacteristicProxy

# -----------------------------------------------------------------------------
# Typing
# -----------------------------------------------------------------------------
_T = TypeVar('_T')
_T2 = TypeVar('_T2', bound=utils.ByteSerializable)
_T3 = TypeVar('_T3', bound=utils.IntConvertible)


# -----------------------------------------------------------------------------
class CharacteristicAdapter(Characteristic, Generic[_T]):
    '''Base class for GATT Characteristic adapters.'''

    def __init__(self, characteristic: Characteristic) -> None:
        super().__init__(
            characteristic.uuid,
            characteristic.properties,
            characteristic.permissions,
            characteristic.value,
            characteristic.descriptors,
        )


# -----------------------------------------------------------------------------
class CharacteristicProxyAdapter(CharacteristicProxy[_T]):
    '''Base class for GATT CharacteristicProxy adapters.'''

    def __init__(self, characteristic_proxy: CharacteristicProxy):
        super().__init__(
            characteristic_proxy.client,
            characteristic_proxy.handle,
            characteristic_proxy.end_group_handle,
            characteristic_proxy.uuid,
            characteristic_proxy.properties,
        )


# -----------------------------------------------------------------------------
class DelegatedCharacteristicAdapter(CharacteristicAdapter[_T]):
    '''
    Adapter that converts bytes values using an encode and/or a decode function.
    '''

    def __init__(
        self,
        characteristic: Characteristic,
        encode: Callable[[_T], bytes] | None = None,
        decode: Callable[[bytes], _T] | None = None,
    ):
        super().__init__(characteristic)
        self.encode = encode
        self.decode = decode

    def encode_value(self, value: _T) -> bytes:
        if self.encode is None:
            raise InvalidOperationError('delegated adapter does not have an encoder')
        return self.encode(value)

    def decode_value(self, value: bytes) -> _T:
        if self.decode is None:
            raise InvalidOperationError('delegate adapter does not have a decoder')
        return self.decode(value)


# -----------------------------------------------------------------------------
class DelegatedCharacteristicProxyAdapter(CharacteristicProxyAdapter[_T]):
    '''
    Adapter that converts bytes values using an encode and a decode function.
    '''

    def __init__(
        self,
        characteristic_proxy: CharacteristicProxy,
        encode: Callable[[_T], bytes] | None = None,
        decode: Callable[[bytes], _T] | None = None,
    ):
        super().__init__(characteristic_proxy)
        self.encode = encode
        self.decode = decode

    def encode_value(self, value: _T) -> bytes:
        if self.encode is None:
            raise InvalidOperationError('delegated adapter does not have an encoder')
        return self.encode(value)

    def decode_value(self, value: bytes) -> _T:
        if self.decode is None:
            raise InvalidOperationError('delegate adapter does not have a decoder')
        return self.decode(value)


# -----------------------------------------------------------------------------
class PackedCharacteristicAdapter(CharacteristicAdapter):
    '''
    Adapter that packs/unpacks characteristic values according to a standard
    Python `struct` format.
    For formats with a single value, the adapted `read_value` and `write_value`
    methods return/accept single values. For formats with multiple values,
    they return/accept a tuple with the same number of elements as is required for
    the format.
    '''

    def __init__(self, characteristic: Characteristic, pack_format: str) -> None:
        super().__init__(characteristic)
        self.struct = struct.Struct(pack_format)

    def pack(self, *values) -> bytes:
        return self.struct.pack(*values)

    def unpack(self, buffer: bytes) -> tuple:
        return self.struct.unpack(buffer)

    def encode_value(self, value: Any) -> bytes:
        return self.pack(*value if isinstance(value, tuple) else (value,))

    def decode_value(self, value: bytes) -> Any:
        unpacked = self.unpack(value)
        return unpacked[0] if len(unpacked) == 1 else unpacked


# -----------------------------------------------------------------------------
class PackedCharacteristicProxyAdapter(CharacteristicProxyAdapter):
    '''
    Adapter that packs/unpacks characteristic values according to a standard
    Python `struct` format.
    For formats with a single value, the adapted `read_value` and `write_value`
    methods return/accept single values. For formats with multiple values,
    they return/accept a tuple with the same number of elements as is required for
    the format.
    '''

    def __init__(self, characteristic_proxy, pack_format):
        super().__init__(characteristic_proxy)
        self.struct = struct.Struct(pack_format)

    def pack(self, *values) -> bytes:
        return self.struct.pack(*values)

    def unpack(self, buffer: bytes) -> tuple:
        return self.struct.unpack(buffer)

    def encode_value(self, value: Any) -> bytes:
        return self.pack(*value if isinstance(value, tuple) else (value,))

    def decode_value(self, value: bytes) -> Any:
        unpacked = self.unpack(value)
        return unpacked[0] if len(unpacked) == 1 else unpacked


# -----------------------------------------------------------------------------
class MappedCharacteristicAdapter(PackedCharacteristicAdapter):
    '''
    Adapter that packs/unpacks characteristic values according to a standard
    Python `struct` format.
    The adapted `read_value` and `write_value` methods return/accept a dictionary which
    is packed/unpacked according to format, with the arguments extracted from the
    dictionary by key, in the same order as they occur in the `keys` parameter.
    '''

    def __init__(
        self, characteristic: Characteristic, pack_format: str, keys: Iterable[str]
    ) -> None:
        super().__init__(characteristic, pack_format)
        self.keys = keys

    # pylint: disable=arguments-differ
    def pack(self, values) -> bytes:
        return super().pack(*(values[key] for key in self.keys))

    def unpack(self, buffer: bytes) -> Any:
        return dict(zip(self.keys, super().unpack(buffer)))


# -----------------------------------------------------------------------------
class MappedCharacteristicProxyAdapter(PackedCharacteristicProxyAdapter):
    '''
    Adapter that packs/unpacks characteristic values according to a standard
    Python `struct` format.
    The adapted `read_value` and `write_value` methods return/accept a dictionary which
    is packed/unpacked according to format, with the arguments extracted from the
    dictionary by key, in the same order as they occur in the `keys` parameter.
    '''

    def __init__(
        self,
        characteristic_proxy: CharacteristicProxy,
        pack_format: str,
        keys: Iterable[str],
    ) -> None:
        super().__init__(characteristic_proxy, pack_format)
        self.keys = keys

    # pylint: disable=arguments-differ
    def pack(self, values) -> bytes:
        return super().pack(*(values[key] for key in self.keys))

    def unpack(self, buffer: bytes) -> Any:
        return dict(zip(self.keys, super().unpack(buffer)))


# -----------------------------------------------------------------------------
class UTF8CharacteristicAdapter(CharacteristicAdapter[str]):
    '''
    Adapter that converts strings to/from bytes using UTF-8 encoding
    '''

    def encode_value(self, value: str) -> bytes:
        return value.encode('utf-8')

    def decode_value(self, value: bytes) -> str:
        return value.decode('utf-8')


# -----------------------------------------------------------------------------
class UTF8CharacteristicProxyAdapter(CharacteristicProxyAdapter[str]):
    '''
    Adapter that converts strings to/from bytes using UTF-8 encoding
    '''

    def encode_value(self, value: str) -> bytes:
        return value.encode('utf-8')

    def decode_value(self, value: bytes) -> str:
        return value.decode('utf-8')


# -----------------------------------------------------------------------------
class SerializableCharacteristicAdapter(CharacteristicAdapter[_T2]):
    '''
    Adapter that converts any class to/from bytes using the class'
    `to_bytes` and `__bytes__` methods, respectively.
    '''

    def __init__(self, characteristic: Characteristic, cls: type[_T2]) -> None:
        super().__init__(characteristic)
        self.cls = cls

    def encode_value(self, value: _T2) -> bytes:
        return bytes(value)

    def decode_value(self, value: bytes) -> _T2:
        return self.cls.from_bytes(value)


# -----------------------------------------------------------------------------
class SerializableCharacteristicProxyAdapter(CharacteristicProxyAdapter[_T2]):
    '''
    Adapter that converts any class to/from bytes using the class'
    `to_bytes` and `__bytes__` methods, respectively.
    '''

    def __init__(
        self, characteristic_proxy: CharacteristicProxy, cls: type[_T2]
    ) -> None:
        super().__init__(characteristic_proxy)
        self.cls = cls

    def encode_value(self, value: _T2) -> bytes:
        return bytes(value)

    def decode_value(self, value: bytes) -> _T2:
        return self.cls.from_bytes(value)


# -----------------------------------------------------------------------------
class EnumCharacteristicAdapter(CharacteristicAdapter[_T3]):
    '''
    Adapter that converts int-enum-like classes to/from bytes using the class'
    `int().to_bytes()` and `from_bytes()` methods, respectively.
    '''

    def __init__(
        self,
        characteristic: Characteristic,
        cls: type[_T3],
        length: int,
        byteorder: Literal['little', 'big'] = 'little',
    ):
        """
        Initialize an instance.

        Params:
          characteristic: the Characteristic to adapt to/from
          cls: the class to/from which to convert integer values
          length: number of bytes used to represent integer values
          byteorder: byte order of the byte representation of integers.
        """
        super().__init__(characteristic)
        self.cls = cls
        self.length = length
        self.byteorder = byteorder

    def encode_value(self, value: _T3) -> bytes:
        return int(value).to_bytes(self.length, self.byteorder)

    def decode_value(self, value: bytes) -> _T3:
        int_value = int.from_bytes(value, self.byteorder)
        return self.cls(int_value)


# -----------------------------------------------------------------------------
class EnumCharacteristicProxyAdapter(CharacteristicProxyAdapter[_T3]):
    '''
    Adapter that converts int-enum-like classes to/from bytes using the class'
    `int().to_bytes()` and `from_bytes()` methods, respectively.
    '''

    def __init__(
        self,
        characteristic_proxy: CharacteristicProxy,
        cls: type[_T3],
        length: int,
        byteorder: Literal['little', 'big'] = 'little',
    ):
        """
        Initialize an instance.

        Params:
          characteristic_proxy: the CharacteristicProxy to adapt to/from
          cls: the class to/from which to convert integer values
          length: number of bytes used to represent integer values
          byteorder: byte order of the byte representation of integers.
        """
        super().__init__(characteristic_proxy)
        self.cls = cls
        self.length = length
        self.byteorder = byteorder

    def encode_value(self, value: _T3) -> bytes:
        return int(value).to_bytes(self.length, self.byteorder)

    def decode_value(self, value: bytes) -> _T3:
        int_value = int.from_bytes(value, self.byteorder)
        return self.cls(int_value)



from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class ControllerPreset(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    DEFAULT: _ClassVar[ControllerPreset]
    LAIRD_BL654: _ClassVar[ControllerPreset]
    CSR_RCK_PTS_DONGLE: _ClassVar[ControllerPreset]
DEFAULT: ControllerPreset
LAIRD_BL654: ControllerPreset
CSR_RCK_PTS_DONGLE: ControllerPreset

class ControllerFeatures(_message.Message):
    __slots__ = ("le_extended_advertising", "le_periodic_advertising", "ll_privacy", "le_2m_phy", "le_coded_phy", "le_connected_isochronous_stream")
    LE_EXTENDED_ADVERTISING_FIELD_NUMBER: _ClassVar[int]
    LE_PERIODIC_ADVERTISING_FIELD_NUMBER: _ClassVar[int]
    LL_PRIVACY_FIELD_NUMBER: _ClassVar[int]
    LE_2M_PHY_FIELD_NUMBER: _ClassVar[int]
    LE_CODED_PHY_FIELD_NUMBER: _ClassVar[int]
    LE_CONNECTED_ISOCHRONOUS_STREAM_FIELD_NUMBER: _ClassVar[int]
    le_extended_advertising: bool
    le_periodic_advertising: bool
    ll_privacy: bool
    le_2m_phy: bool
    le_coded_phy: bool
    le_connected_isochronous_stream: bool
    def __init__(self, le_extended_advertising: bool = ..., le_periodic_advertising: bool = ..., ll_privacy: bool = ..., le_2m_phy: bool = ..., le_coded_phy: bool = ..., le_connected_isochronous_stream: bool = ...) -> None: ...

class ControllerQuirks(_message.Message):
    __slots__ = ("send_acl_data_before_connection_complete", "has_default_random_address", "hardware_error_before_reset")
    SEND_ACL_DATA_BEFORE_CONNECTION_COMPLETE_FIELD_NUMBER: _ClassVar[int]
    HAS_DEFAULT_RANDOM_ADDRESS_FIELD_NUMBER: _ClassVar[int]
    HARDWARE_ERROR_BEFORE_RESET_FIELD_NUMBER: _ClassVar[int]
    send_acl_data_before_connection_complete: bool
    has_default_random_address: bool
    hardware_error_before_reset: bool
    def __init__(self, send_acl_data_before_connection_complete: bool = ..., has_default_random_address: bool = ..., hardware_error_before_reset: bool = ...) -> None: ...

class VendorFeatures(_message.Message):
    __slots__ = ("csr", "android")
    CSR_FIELD_NUMBER: _ClassVar[int]
    ANDROID_FIELD_NUMBER: _ClassVar[int]
    csr: bool
    android: bool
    def __init__(self, csr: bool = ..., android: bool = ...) -> None: ...

class Controller(_message.Message):
    __slots__ = ("preset", "features", "quirks", "strict", "vendor")
    PRESET_FIELD_NUMBER: _ClassVar[int]
    FEATURES_FIELD_NUMBER: _ClassVar[int]
    QUIRKS_FIELD_NUMBER: _ClassVar[int]
    STRICT_FIELD_NUMBER: _ClassVar[int]
    VENDOR_FIELD_NUMBER: _ClassVar[int]
    preset: ControllerPreset
    features: ControllerFeatures
    quirks: ControllerQuirks
    strict: bool
    vendor: VendorFeatures
    def __init__(self, preset: _Optional[_Union[ControllerPreset, str]] = ..., features: _Optional[_Union[ControllerFeatures, _Mapping]] = ..., quirks: _Optional[_Union[ControllerQuirks, _Mapping]] = ..., strict: bool = ..., vendor: _Optional[_Union[VendorFeatures, _Mapping]] = ...) -> None: ...

class TcpServer(_message.Message):
    __slots__ = ("tcp_port", "configuration")
    TCP_PORT_FIELD_NUMBER: _ClassVar[int]
    CONFIGURATION_FIELD_NUMBER: _ClassVar[int]
    tcp_port: int
    configuration: Controller
    def __init__(self, tcp_port: _Optional[int] = ..., configuration: _Optional[_Union[Controller, _Mapping]] = ...) -> None: ...

class Configuration(_message.Message):
    __slots__ = ("tcp_server",)
    TCP_SERVER_FIELD_NUMBER: _ClassVar[int]
    tcp_server: _containers.RepeatedCompositeFieldContainer[TcpServer]
    def __init__(self, tcp_server: _Optional[_Iterable[_Union[TcpServer, _Mapping]]] = ...) -> None: ...



from . import emulated_bluetooth_packets_pb2 as _emulated_bluetooth_packets_pb2
from . import emulated_bluetooth_device_pb2 as _emulated_bluetooth_device_pb2
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class RawData(_message.Message):
    __slots__ = ["packet"]
    PACKET_FIELD_NUMBER: _ClassVar[int]
    packet: bytes
    def __init__(self, packet: _Optional[bytes] = ...) -> None: ...

class RegistrationStatus(_message.Message):
    __slots__ = ["callback_device_id"]
    CALLBACK_DEVICE_ID_FIELD_NUMBER: _ClassVar[int]
    callback_device_id: _emulated_bluetooth_device_pb2.CallbackIdentifier
    def __init__(self, callback_device_id: _Optional[_Union[_emulated_bluetooth_device_pb2.CallbackIdentifier, _Mapping]] = ...) -> None: ...



# Copyright 2021-2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------
import asyncio
import sys

import bumble.logging
from bumble.core import (
    BT_ADVANCED_AUDIO_DISTRIBUTION_SERVICE,
    BT_AUDIO_SINK_SERVICE,
    BT_AVDTP_PROTOCOL_ID,
    BT_L2CAP_PROTOCOL_ID,
)
from bumble.device import Device
from bumble.sdp import (
    SDP_BLUETOOTH_PROFILE_DESCRIPTOR_LIST_ATTRIBUTE_ID,
    SDP_BROWSE_GROUP_LIST_ATTRIBUTE_ID,
    SDP_PROTOCOL_DESCRIPTOR_LIST_ATTRIBUTE_ID,
    SDP_PUBLIC_BROWSE_ROOT,
    SDP_SERVICE_CLASS_ID_LIST_ATTRIBUTE_ID,
    SDP_SERVICE_RECORD_HANDLE_ATTRIBUTE_ID,
    DataElement,
    ServiceAttribute,
)
from bumble.transport import open_transport

# -----------------------------------------------------------------------------
SDP_SERVICE_RECORDS = {
    0x00010001: [
        ServiceAttribute(
            SDP_SERVICE_RECORD_HANDLE_ATTRIBUTE_ID,
            DataElement.unsigned_integer_32(0x00010001),
        ),
        ServiceAttribute(
            SDP_BROWSE_GROUP_LIST_ATTRIBUTE_ID,
            DataElement.sequence([DataElement.uuid(SDP_PUBLIC_BROWSE_ROOT)]),
        ),
        ServiceAttribute(
            SDP_SERVICE_CLASS_ID_LIST_ATTRIBUTE_ID,
            DataElement.sequence([DataElement.uuid(BT_AUDIO_SINK_SERVICE)]),
        ),
        ServiceAttribute(
            SDP_PROTOCOL_DESCRIPTOR_LIST_ATTRIBUTE_ID,
            DataElement.sequence(
                [
                    DataElement.sequence(
                        [
                            DataElement.uuid(BT_L2CAP_PROTOCOL_ID),
                            DataElement.unsigned_integer_16(25),
                        ]
                    ),
                    DataElement.sequence(
                        [
                            DataElement.uuid(BT_AVDTP_PROTOCOL_ID),
                            DataElement.unsigned_integer_16(256),
                        ]
                    ),
                ]
            ),
        ),
        ServiceAttribute(
            SDP_BLUETOOTH_PROFILE_DESCRIPTOR_LIST_ATTRIBUTE_ID,
            DataElement.sequence(
                [
                    DataElement.sequence(
                        [
                            DataElement.uuid(BT_ADVANCED_AUDIO_DISTRIBUTION_SERVICE),
                            DataElement.unsigned_integer_16(256),
                        ]
                    )
                ]
            ),
        ),
    ]
}


# -----------------------------------------------------------------------------
async def main() -> None:
    if len(sys.argv) < 3:
        print('Usage: run_classic_discoverable.py <device-config> <transport-spec>')
        print('example: run_classic_discoverable.py classic1.json usb:04b4:f901')
        return

    print('<<< connecting to HCI...')
    async with await open_transport(sys.argv[2]) as hci_transport:
        print('<<< connected')

        # Create a device
        device = Device.from_config_file_with_hci(
            sys.argv[1], hci_transport.source, hci_transport.sink
        )
        device.classic_enabled = True
        device.sdp_service_records = SDP_SERVICE_RECORDS
        await device.power_on()

        # Start being discoverable and connectable
        await device.set_discoverable(True)
        await device.set_connectable(True)

        await hci_transport.source.terminated


# -----------------------------------------------------------------------------
bumble.logging.setup_basic_logging('DEBUG')
asyncio.run(main())



# Copyright 2021-2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------
import asyncio
import logging
import os
import struct
from collections.abc import Awaitable

import pytest

from bumble import a2dp, avdtp
from bumble.controller import Controller
from bumble.core import PhysicalTransport
from bumble.device import Device
from bumble.host import Host
from bumble.link import LocalLink
from bumble.rtp import MediaPacket
from bumble.transport.common import AsyncPipeSink

# -----------------------------------------------------------------------------
# Logging
# -----------------------------------------------------------------------------
logger = logging.getLogger(__name__)


# -----------------------------------------------------------------------------
class TwoDevices:
    def __init__(self):
        self.connections = [None, None]

        addresses = ['F0:F1:F2:F3:F4:F5', 'F5:F4:F3:F2:F1:F0']
        self.link = LocalLink()
        self.controllers = [
            Controller('C1', link=self.link, public_address=addresses[0]),
            Controller('C2', link=self.link, public_address=addresses[1]),
        ]
        self.devices = [
            Device(
                address=addresses[0],
                host=Host(self.controllers[0], AsyncPipeSink(self.controllers[0])),
            ),
            Device(
                address=addresses[1],
                host=Host(self.controllers[1], AsyncPipeSink(self.controllers[1])),
            ),
        ]

        self.paired = [None, None]

    def on_connection(self, which, connection):
        self.connections[which] = connection

    def on_paired(self, which, keys):
        self.paired[which] = keys


# -----------------------------------------------------------------------------
class Data:
    pointer: int = 0
    data: bytes

    def __init__(self, data: bytes):
        self.data = data

    async def read(self, length: int) -> Awaitable[bytes]:
        def generate_read():
            end = min(self.pointer + length, len(self.data))
            chunk = self.data[self.pointer : end]
            self.pointer = end
            return chunk

        return generate_read()


# -----------------------------------------------------------------------------
@pytest.mark.asyncio
async def test_self_connection():
    # Create two devices, each with a controller, attached to the same link
    two_devices = TwoDevices()

    # Attach listeners
    two_devices.devices[0].on(
        'connection', lambda connection: two_devices.on_connection(0, connection)
    )
    two_devices.devices[1].on(
        'connection', lambda connection: two_devices.on_connection(1, connection)
    )

    # Enable Classic connections
    two_devices.devices[0].classic_enabled = True
    two_devices.devices[1].classic_enabled = True

    # Start
    await two_devices.devices[0].power_on()
    await two_devices.devices[1].power_on()

    # Connect the two devices
    await asyncio.gather(
        two_devices.devices[0].connect(
            two_devices.devices[1].public_address, transport=PhysicalTransport.BR_EDR
        ),
        two_devices.devices[1].accept(two_devices.devices[0].public_address),
    )

    # Check the post conditions
    assert two_devices.connections[0] is not None
    assert two_devices.connections[1] is not None


# -----------------------------------------------------------------------------
def source_codec_capabilities():
    return avdtp.MediaCodecCapabilities(
        media_type=avdtp.MediaType.AUDIO,
        media_codec_type=a2dp.CodecType.SBC,
        media_codec_information=a2dp.SbcMediaCodecInformation(
            sampling_frequency=a2dp.SbcMediaCodecInformation.SamplingFrequency.SF_44100,
            channel_mode=a2dp.SbcMediaCodecInformation.ChannelMode.JOINT_STEREO,
            block_length=a2dp.SbcMediaCodecInformation.BlockLength.BL_16,
            subbands=a2dp.SbcMediaCodecInformation.Subbands.S_8,
            allocation_method=a2dp.SbcMediaCodecInformation.AllocationMethod.LOUDNESS,
            minimum_bitpool_value=2,
            maximum_bitpool_value=53,
        ),
    )


# -----------------------------------------------------------------------------
def sink_codec_capabilities():
    return avdtp.MediaCodecCapabilities(
        media_type=avdtp.MediaType.AUDIO,
        media_codec_type=a2dp.CodecType.SBC,
        media_codec_information=a2dp.SbcMediaCodecInformation(
            sampling_frequency=a2dp.SbcMediaCodecInformation.SamplingFrequency.SF_48000
            | a2dp.SbcMediaCodecInformation.SamplingFrequency.SF_44100
            | a2dp.SbcMediaCodecInformation.SamplingFrequency.SF_32000
            | a2dp.SbcMediaCodecInformation.SamplingFrequency.SF_16000,
            channel_mode=a2dp.SbcMediaCodecInformation.ChannelMode.MONO
            | a2dp.SbcMediaCodecInformation.ChannelMode.DUAL_CHANNEL
            | a2dp.SbcMediaCodecInformation.ChannelMode.STEREO
            | a2dp.SbcMediaCodecInformation.ChannelMode.JOINT_STEREO,
            block_length=a2dp.SbcMediaCodecInformation.BlockLength.BL_4
            | a2dp.SbcMediaCodecInformation.BlockLength.BL_8
            | a2dp.SbcMediaCodecInformation.BlockLength.BL_12
            | a2dp.SbcMediaCodecInformation.BlockLength.BL_16,
            subbands=a2dp.SbcMediaCodecInformation.Subbands.S_4
            | a2dp.SbcMediaCodecInformation.Subbands.S_8,
            allocation_method=a2dp.SbcMediaCodecInformation.AllocationMethod.LOUDNESS
            | a2dp.SbcMediaCodecInformation.AllocationMethod.SNR,
            minimum_bitpool_value=2,
            maximum_bitpool_value=53,
        ),
    )


# -----------------------------------------------------------------------------
@pytest.mark.asyncio
async def test_source_sink_1():
    two_devices = TwoDevices()
    # Enable Classic connections
    two_devices.devices[0].classic_enabled = True
    two_devices.devices[1].classic_enabled = True
    await two_devices.devices[0].power_on()
    await two_devices.devices[1].power_on()

    def on_rtp_packet(packet):
        rtp_packets.append(packet)
        if len(rtp_packets) == rtp_packets_expected:
            rtp_packets_fully_received.set_result(None)

    sink = None

    def on_avdtp_connection(server):
        nonlocal sink
        sink = server.add_sink(sink_codec_capabilities())
        sink.on('rtp_packet', on_rtp_packet)

    # Create a listener to wait for AVDTP connections
    listener = avdtp.Listener.for_device(two_devices.devices[1])
    listener.on('connection', on_avdtp_connection)

    async def make_connection():
        connections = await asyncio.gather(
            two_devices.devices[0].connect(
                two_devices.devices[1].public_address, PhysicalTransport.BR_EDR
            ),
            two_devices.devices[1].accept(two_devices.devices[0].public_address),
        )
        return connections[0]

    connection = await make_connection()
    client = await avdtp.Protocol.connect(connection)
    endpoints = await client.discover_remote_endpoints()
    assert len(endpoints) == 1
    remote_sink = list(endpoints)[0]
    assert remote_sink.in_use == 0
    assert remote_sink.media_type == avdtp.MediaType.AUDIO
    assert remote_sink.tsep == avdtp.StreamEndPointType.SNK

    async def generate_packets(packet_count):
        sequence_number = 0
        timestamp = 0
        for i in range(packet_count):
            payload = bytes([sequence_number % 256])
            packet = MediaPacket(
                2, 0, 0, 0, sequence_number, timestamp, 0, [], 96, payload
            )
            packet.timestamp_seconds = timestamp / 44100
            timestamp += 10
            sequence_number += 1
            yield packet

    # Send packets using a pump object
    rtp_packets_fully_received = asyncio.get_running_loop().create_future()
    rtp_packets_expected = 3
    rtp_packets = []
    pump = avdtp.MediaPacketPump(generate_packets(3))
    source = client.add_source(source_codec_capabilities(), pump)
    stream = await client.create_stream(source, remote_sink)
    await stream.start()
    assert stream.state == avdtp.State.STREAMING
    assert stream.local_endpoint.in_use == 1
    assert stream.rtp_channel is not None
    assert sink.in_use == 1
    assert sink.stream is not None
    assert sink.stream.state == avdtp.State.STREAMING
    await rtp_packets_fully_received

    await stream.close()
    assert stream.rtp_channel is None
    assert source.in_use == 0
    assert source.stream.state == avdtp.State.IDLE
    assert sink.in_use == 0
    assert sink.stream.state == avdtp.State.IDLE

    # Send packets manually
    rtp_packets_fully_received = asyncio.get_running_loop().create_future()
    rtp_packets_expected = 3
    rtp_packets = []
    source_packets = [
        MediaPacket(2, 0, 0, 0, i, i * 10, 0, [], 96, bytes([i])) for i in range(3)
    ]
    source = client.add_source(source_codec_capabilities(), None)
    stream = await client.create_stream(source, remote_sink)
    await stream.start()
    assert stream.state == avdtp.State.STREAMING
    assert stream.local_endpoint.in_use == 1
    assert stream.rtp_channel is not None
    assert sink.in_use == 1
    assert sink.stream is not None
    assert sink.stream.state == avdtp.State.STREAMING

    stream.send_media_packet(source_packets[0])
    stream.send_media_packet(source_packets[1])
    stream.send_media_packet(source_packets[2])

    await stream.close()
    assert stream.rtp_channel is None
    assert len(rtp_packets) == 3
    assert source.in_use == 0
    assert source.stream.state == avdtp.State.IDLE
    assert sink.in_use == 0
    assert sink.stream.state == avdtp.State.IDLE


# -----------------------------------------------------------------------------
def test_sbc_codec_specific_information():
    sbc_info = a2dp.SbcMediaCodecInformation.from_bytes(bytes.fromhex("3fff0235"))
    assert (
        sbc_info.sampling_frequency
        == a2dp.SbcMediaCodecInformation.SamplingFrequency.SF_44100
        | a2dp.SbcMediaCodecInformation.SamplingFrequency.SF_48000
    )
    assert (
        sbc_info.channel_mode
        == a2dp.SbcMediaCodecInformation.ChannelMode.MONO
        | a2dp.SbcMediaCodecInformation.ChannelMode.DUAL_CHANNEL
        | a2dp.SbcMediaCodecInformation.ChannelMode.STEREO
        | a2dp.SbcMediaCodecInformation.ChannelMode.JOINT_STEREO
    )
    assert (
        sbc_info.block_length
        == a2dp.SbcMediaCodecInformation.BlockLength.BL_4
        | a2dp.SbcMediaCodecInformation.BlockLength.BL_8
        | a2dp.SbcMediaCodecInformation.BlockLength.BL_12
        | a2dp.SbcMediaCodecInformation.BlockLength.BL_16
    )
    assert (
        sbc_info.subbands
        == a2dp.SbcMediaCodecInformation.Subbands.S_4
        | a2dp.SbcMediaCodecInformation.Subbands.S_8
    )
    assert (
        sbc_info.allocation_method
        == a2dp.SbcMediaCodecInformation.AllocationMethod.SNR
        | a2dp.SbcMediaCodecInformation.AllocationMethod.LOUDNESS
    )
    assert sbc_info.minimum_bitpool_value == 2
    assert sbc_info.maximum_bitpool_value == 53

    sbc_info2 = a2dp.SbcMediaCodecInformation(
        a2dp.SbcMediaCodecInformation.SamplingFrequency.SF_44100
        | a2dp.SbcMediaCodecInformation.SamplingFrequency.SF_48000,
        a2dp.SbcMediaCodecInformation.ChannelMode.MONO
        | a2dp.SbcMediaCodecInformation.ChannelMode.DUAL_CHANNEL
        | a2dp.SbcMediaCodecInformation.ChannelMode.STEREO
        | a2dp.SbcMediaCodecInformation.ChannelMode.JOINT_STEREO,
        a2dp.SbcMediaCodecInformation.BlockLength.BL_4
        | a2dp.SbcMediaCodecInformation.BlockLength.BL_8
        | a2dp.SbcMediaCodecInformation.BlockLength.BL_12
        | a2dp.SbcMediaCodecInformation.BlockLength.BL_16,
        a2dp.SbcMediaCodecInformation.Subbands.S_4
        | a2dp.SbcMediaCodecInformation.Subbands.S_8,
        a2dp.SbcMediaCodecInformation.AllocationMethod.SNR
        | a2dp.SbcMediaCodecInformation.AllocationMethod.LOUDNESS,
        2,
        53,
    )
    assert sbc_info == sbc_info2
    assert bytes(sbc_info2) == bytes.fromhex("3fff0235")


# -----------------------------------------------------------------------------
def test_aac_codec_specific_information():
    aac_info = a2dp.AacMediaCodecInformation.from_bytes(bytes.fromhex("f0018c83e800"))
    assert (
        aac_info.object_type
        == a2dp.AacMediaCodecInformation.ObjectType.MPEG_2_AAC_LC
        | a2dp.AacMediaCodecInformation.ObjectType.MPEG_4_AAC_LC
        | a2dp.AacMediaCodecInformation.ObjectType.MPEG_4_AAC_LTP
        | a2dp.AacMediaCodecInformation.ObjectType.MPEG_4_AAC_SCALABLE
    )
    assert (
        aac_info.sampling_frequency
        == a2dp.AacMediaCodecInformation.SamplingFrequency.SF_44100
        | a2dp.AacMediaCodecInformation.SamplingFrequency.SF_48000
    )
    assert (
        aac_info.channels
        == a2dp.AacMediaCodecInformation.Channels.MONO
        | a2dp.AacMediaCodecInformation.Channels.STEREO
    )
    assert aac_info.vbr == 1
    assert aac_info.bitrate == 256000

    aac_info2 = a2dp.AacMediaCodecInformation(
        a2dp.AacMediaCodecInformation.ObjectType.MPEG_2_AAC_LC
        | a2dp.AacMediaCodecInformation.ObjectType.MPEG_4_AAC_LC
        | a2dp.AacMediaCodecInformation.ObjectType.MPEG_4_AAC_LTP
        | a2dp.AacMediaCodecInformation.ObjectType.MPEG_4_AAC_SCALABLE,
        a2dp.AacMediaCodecInformation.SamplingFrequency.SF_44100
        | a2dp.AacMediaCodecInformation.SamplingFrequency.SF_48000,
        a2dp.AacMediaCodecInformation.Channels.MONO
        | a2dp.AacMediaCodecInformation.Channels.STEREO,
        1,
        256000,
    )
    assert aac_info == aac_info2
    assert bytes(aac_info2) == bytes.fromhex("f0018c83e800")


# -----------------------------------------------------------------------------
def test_opus_codec_specific_information():
    opus_info = a2dp.OpusMediaCodecInformation.from_bytes(bytes([0x92]))
    assert opus_info.vendor_id == a2dp.OpusMediaCodecInformation.VENDOR_ID
    assert opus_info.codec_id == a2dp.OpusMediaCodecInformation.CODEC_ID
    assert opus_info.frame_size == a2dp.OpusMediaCodecInformation.FrameSize.FS_20MS
    assert opus_info.channel_mode == a2dp.OpusMediaCodecInformation.ChannelMode.STEREO
    assert (
        opus_info.sampling_frequency
        == a2dp.OpusMediaCodecInformation.SamplingFrequency.SF_48000
    )

    opus_info2 = a2dp.OpusMediaCodecInformation(
        a2dp.OpusMediaCodecInformation.ChannelMode.STEREO,
        a2dp.OpusMediaCodecInformation.FrameSize.FS_20MS,
        a2dp.OpusMediaCodecInformation.SamplingFrequency.SF_48000,
    )
    assert opus_info2 == opus_info
    assert opus_info2.value == bytes([0x92])


# -----------------------------------------------------------------------------
async def test_sbc_parser():
    header = b'\x9c\x80\x08\x00'
    payload = b'\x00\x00\x00\x00\x00\x00'
    data = Data(header + payload)

    parser = a2dp.SbcParser(data.read)
    async for frame in parser.frames:
        assert frame.sampling_frequency == 44100
        assert frame.block_count == 4
        assert frame.channel_mode == 0
        assert frame.allocation_method == 0
        assert frame.subband_count == 4
        assert frame.bitpool == 8
        assert frame.payload == header + payload


# -----------------------------------------------------------------------------
async def test_sbc_packet_source():
    header = b'\x9c\x80\x08\x00'
    payload = b'\x00\x00\x00\x00\x00\x00'
    data = Data((header + payload) * 2)

    packet_source = a2dp.SbcPacketSource(data.read, 23)
    async for packet in packet_source.packets:
        assert packet.sequence_number == 0
        assert packet.timestamp == 0
        assert packet.payload == b'\x01' + header + payload


# -----------------------------------------------------------------------------
async def test_aac_parser():
    header = b'\xff\xf0\x10\x00\x01\xa0\x00'
    payload = b'\x00\x00\x00\x00\x00\x00'
    data = Data(header + payload)

    parser = a2dp.AacParser(data.read)
    async for frame in parser.frames:
        assert frame.profile == a2dp.AacFrame.Profile.MAIN
        assert frame.sampling_frequency == 44100
        assert frame.channel_configuration == 0
        assert frame.payload == payload


# -----------------------------------------------------------------------------
async def test_aac_packet_source():
    header = b'\xff\xf0\x10\x00\x01\xa0\x00'
    payload = b'\x00\x00\x00\x00\x00\x00'
    data = Data(header + payload)

    packet_source = a2dp.AacPacketSource(data.read, 0)
    async for packet in packet_source.packets:
        assert packet.sequence_number == 0
        assert packet.timestamp == 0
        assert packet.payload == b' \x00\x12\x00\x00\x000\x00\x00\x00\x00\x00\x00'


# -----------------------------------------------------------------------------
async def test_opus_parser():
    packed_header_data_revised = struct.pack(
        "<QIIIB",
        0,  # granule_position
        2,  # bitstream_serial_number
        2,  # page_sequence_number
        0,  # crc_checksum
        3,  # page_segments
    )

    first_page_header_revised = (
        b'OggS'  # Capture pattern
        + b'\x00'  # Version
        + b'\x02'  # Header type
        + packed_header_data_revised
    )

    segment_table_revised = b'\x0a\x08\x0a'

    opus_head_packet_data = b'OpusHead' + b'\x00' + b'\x00'
    opus_tags_packet_data = b'OpusTags'
    audio_data_packet = b'0123456789'

    data = Data(
        first_page_header_revised
        + segment_table_revised
        + opus_head_packet_data
        + opus_tags_packet_data
        + audio_data_packet
    )

    parser = a2dp.OpusParser(data.read)
    async for packet in parser.packets:
        assert packet.channel_mode == a2dp.OpusPacket.ChannelMode.STEREO
        assert packet.payload == audio_data_packet


# -----------------------------------------------------------------------------
async def test_opus_packet_source():
    packed_header_data_revised = struct.pack(
        "<QIIIB",
        0,  # granule_position
        2,  # bitstream_serial_number
        2,  # page_sequence_number
        0,  # crc_checksum
        3,  # page_segments
    )

    first_page_header_revised = (
        b'OggS'  # Capture pattern
        + b'\x00'  # Version
        + b'\x02'  # Header type
        + packed_header_data_revised
    )

    segment_table_revised = b'\x0a\x08\x0a'

    opus_head_packet_data = b'OpusHead' + b'\x00' + b'\x00'
    opus_tags_packet_data = b'OpusTags'
    audio_data_packet = b'0123456789'

    data = Data(
        first_page_header_revised
        + segment_table_revised
        + opus_head_packet_data
        + opus_tags_packet_data
        + audio_data_packet
    )

    parser = a2dp.OpusPacketSource(data.read, 0)
    async for packet in parser.packets:
        assert packet.sequence_number == 0
        assert packet.timestamp == 0
        assert packet.payload == b'\x01' + audio_data_packet


# -----------------------------------------------------------------------------
async def async_main():
    test_sbc_codec_specific_information()
    test_aac_codec_specific_information()
    test_opus_codec_specific_information()
    await test_self_connection()
    await test_source_sink_1()
    test_sbc_parser()
    test_sbc_packet_source()
    test_aac_parser()
    test_aac_packet_source()
    test_opus_parser()
    test_opus_packet_source()


# -----------------------------------------------------------------------------
if __name__ == '__main__':
    logging.basicConfig(level=os.environ.get('BUMBLE_LOGLEVEL', 'INFO').upper())
    asyncio.run(async_main())



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Apple Notification Center Service (ANCS).
"""

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------
from __future__ import annotations

import asyncio
import dataclasses
import datetime
import enum
import logging
import struct
from collections.abc import Sequence

from bumble import utils
from bumble.att import ATT_Error
from bumble.device import Peer
from bumble.gatt import (
    GATT_ANCS_CONTROL_POINT_CHARACTERISTIC,
    GATT_ANCS_DATA_SOURCE_CHARACTERISTIC,
    GATT_ANCS_NOTIFICATION_SOURCE_CHARACTERISTIC,
    GATT_ANCS_SERVICE,
    Characteristic,
    TemplateService,
)
from bumble.gatt_adapters import SerializableCharacteristicProxyAdapter
from bumble.gatt_client import CharacteristicProxy, ProfileServiceProxy, ServiceProxy

# -----------------------------------------------------------------------------
# Constants
# -----------------------------------------------------------------------------
_DEFAULT_ATTRIBUTE_MAX_LENGTH = 65535


# -----------------------------------------------------------------------------
# Logging
# -----------------------------------------------------------------------------
logger = logging.getLogger(__name__)


# -----------------------------------------------------------------------------
# Protocol
# -----------------------------------------------------------------------------
class ActionId(utils.OpenIntEnum):
    POSITIVE = 0
    NEGATIVE = 1


class AppAttributeId(utils.OpenIntEnum):
    DISPLAY_NAME = 0


class CategoryId(utils.OpenIntEnum):
    OTHER = 0
    INCOMING_CALL = 1
    MISSED_CALL = 2
    VOICEMAIL = 3
    SOCIAL = 4
    SCHEDULE = 5
    EMAIL = 6
    NEWS = 7
    HEALTH_AND_FITNESS = 8
    BUSINESS_AND_FINANCE = 9
    LOCATION = 10
    ENTERTAINMENT = 11


class CommandId(utils.OpenIntEnum):
    GET_NOTIFICATION_ATTRIBUTES = 0
    GET_APP_ATTRIBUTES = 1
    PERFORM_NOTIFICATION_ACTION = 2


class EventId(utils.OpenIntEnum):
    NOTIFICATION_ADDED = 0
    NOTIFICATION_MODIFIED = 1
    NOTIFICATION_REMOVED = 2


class EventFlags(enum.IntFlag):
    SILENT = 1 << 0
    IMPORTANT = 1 << 1
    PRE_EXISTING = 1 << 2
    POSITIVE_ACTION = 1 << 3
    NEGATIVE_ACTION = 1 << 4


class NotificationAttributeId(utils.OpenIntEnum):
    APP_IDENTIFIER = 0
    TITLE = 1
    SUBTITLE = 2
    MESSAGE = 3
    MESSAGE_SIZE = 4
    DATE = 5
    POSITIVE_ACTION_LABEL = 6
    NEGATIVE_ACTION_LABEL = 7


@dataclasses.dataclass
class NotificationAttribute:
    attribute_id: NotificationAttributeId
    value: str | int | datetime.datetime


@dataclasses.dataclass
class AppAttribute:
    attribute_id: AppAttributeId
    value: str


@dataclasses.dataclass
class Notification:
    event_id: EventId
    event_flags: EventFlags
    category_id: CategoryId
    category_count: int
    notification_uid: int

    @classmethod
    def from_bytes(cls, data: bytes) -> Notification:
        return cls(
            event_id=EventId(data[0]),
            event_flags=EventFlags(data[1]),
            category_id=CategoryId(data[2]),
            category_count=data[3],
            notification_uid=int.from_bytes(data[4:8], 'little'),
        )

    def __bytes__(self) -> bytes:
        return struct.pack(
            "<BBBBI",
            self.event_id,
            self.event_flags,
            self.category_id,
            self.category_count,
            self.notification_uid,
        )


class ErrorCode(utils.OpenIntEnum):
    UNKNOWN_COMMAND = 0xA0
    INVALID_COMMAND = 0xA1
    INVALID_PARAMETER = 0xA2
    ACTION_FAILED = 0xA3


class ProtocolError(Exception):
    pass


class CommandError(Exception):
    def __init__(self, error_code: ErrorCode) -> None:
        self.error_code = error_code

    def __str__(self) -> str:
        return f"CommandError(error_code={self.error_code.name})"


# -----------------------------------------------------------------------------
# GATT Server-side
# -----------------------------------------------------------------------------
class Ancs(TemplateService):
    UUID = GATT_ANCS_SERVICE

    notification_source_characteristic: Characteristic
    data_source_characteristic: Characteristic
    control_point_characteristic: Characteristic

    def __init__(self) -> None:
        # TODO not the final implementation
        self.notification_source_characteristic = Characteristic(
            GATT_ANCS_NOTIFICATION_SOURCE_CHARACTERISTIC,
            Characteristic.Properties.NOTIFY,
            Characteristic.Permissions.READABLE,
        )

        # TODO not the final implementation
        self.data_source_characteristic = Characteristic(
            GATT_ANCS_DATA_SOURCE_CHARACTERISTIC,
            Characteristic.Properties.NOTIFY,
            Characteristic.Permissions.READABLE,
        )

        # TODO not the final implementation
        self.control_point_characteristic = Characteristic(
            GATT_ANCS_CONTROL_POINT_CHARACTERISTIC,
            Characteristic.Properties.WRITE,
            Characteristic.Permissions.WRITEABLE,
        )

        super().__init__(
            [
                self.notification_source_characteristic,
                self.data_source_characteristic,
                self.control_point_characteristic,
            ]
        )


# -----------------------------------------------------------------------------
# GATT Client-side
# -----------------------------------------------------------------------------
class AncsProxy(ProfileServiceProxy):
    SERVICE_CLASS = Ancs

    notification_source: CharacteristicProxy[Notification]
    data_source: CharacteristicProxy
    control_point: CharacteristicProxy[bytes]

    def __init__(self, service_proxy: ServiceProxy):
        self.notification_source = SerializableCharacteristicProxyAdapter(
            service_proxy.get_required_characteristic_by_uuid(
                GATT_ANCS_NOTIFICATION_SOURCE_CHARACTERISTIC
            ),
            Notification,
        )

        self.data_source = service_proxy.get_required_characteristic_by_uuid(
            GATT_ANCS_DATA_SOURCE_CHARACTERISTIC
        )

        self.control_point = service_proxy.get_required_characteristic_by_uuid(
            GATT_ANCS_CONTROL_POINT_CHARACTERISTIC
        )


class AncsClient(utils.EventEmitter):
    _expected_response_command_id: CommandId | None
    _expected_response_notification_uid: int | None
    _expected_response_app_identifier: str | None
    _expected_app_identifier: str | None
    _expected_response_tuples: int
    _response_accumulator: bytes

    EVENT_NOTIFICATION = "notification"

    def __init__(self, ancs_proxy: AncsProxy) -> None:
        super().__init__()
        self._ancs_proxy = ancs_proxy
        self._command_semaphore = asyncio.Semaphore()
        self._response: asyncio.Future | None = None
        self._reset_response()
        self._started = False

    @classmethod
    async def for_peer(cls, peer: Peer) -> AncsClient | None:
        ancs_proxy = await peer.discover_service_and_create_proxy(AncsProxy)
        if ancs_proxy is None:
            return None
        return cls(ancs_proxy)

    async def start(self) -> None:
        await self._ancs_proxy.notification_source.subscribe(self._on_notification)
        await self._ancs_proxy.data_source.subscribe(self._on_data)
        self._started = True

    async def stop(self) -> None:
        await self._ancs_proxy.notification_source.unsubscribe(self._on_notification)
        await self._ancs_proxy.data_source.unsubscribe(self._on_data)
        self._started = False

    def _reset_response(self) -> None:
        self._expected_response_command_id = None
        self._expected_response_notification_uid = None
        self._expected_app_identifier = None
        self._expected_response_tuples = 0
        self._response_accumulator = b""

    def _on_notification(self, notification: Notification) -> None:
        logger.debug(f"ANCS NOTIFICATION: {notification}")
        self.emit(self.EVENT_NOTIFICATION, notification)

    def _on_data(self, data: bytes) -> None:
        logger.debug(f"ANCS DATA: {data.hex()}")

        if not self._response:
            logger.warning("received unexpected data, discarding")
            return

        self._response_accumulator += data

        # Try to parse the accumulated data until we have all we need.
        if not self._response_accumulator:
            logger.warning("empty data from data source")
            return

        command_id = self._response_accumulator[0]
        if command_id != self._expected_response_command_id:
            logger.warning(
                "unexpected response command id: "
                f"expected {self._expected_response_command_id} "
                f"but got {command_id}"
            )
            self._reset_response()
            if not self._response.done():
                self._response.set_exception(ProtocolError())

        if len(self._response_accumulator) < 5:
            # Not enough data yet.
            return

        attributes: list[NotificationAttribute | AppAttribute] = []

        if command_id == CommandId.GET_NOTIFICATION_ATTRIBUTES:
            (notification_uid,) = struct.unpack_from(
                "<I", self._response_accumulator, 1
            )
            if notification_uid != self._expected_response_notification_uid:
                logger.warning(
                    "unexpected response notification uid: "
                    f"expected {self._expected_response_notification_uid} "
                    f"but got {notification_uid}"
                )
                self._reset_response()
                if not self._response.done():
                    self._response.set_exception(ProtocolError())

            attribute_data = self._response_accumulator[5:]
            while len(attribute_data) >= 3:
                attribute_id, attribute_data_length = struct.unpack_from(
                    "<BH", attribute_data, 0
                )
                if len(attribute_data) < 3 + attribute_data_length:
                    return
                str_value = attribute_data[3 : 3 + attribute_data_length].decode(
                    "utf-8"
                )
                value: str | int | datetime.datetime
                if attribute_id == NotificationAttributeId.MESSAGE_SIZE:
                    value = int(str_value)
                elif attribute_id == NotificationAttributeId.DATE:
                    year = int(str_value[:4])
                    month = int(str_value[4:6])
                    day = int(str_value[6:8])
                    hour = int(str_value[9:11])
                    minute = int(str_value[11:13])
                    second = int(str_value[13:15])
                    value = datetime.datetime(year, month, day, hour, minute, second)
                else:
                    value = str_value
                attributes.append(
                    NotificationAttribute(NotificationAttributeId(attribute_id), value)
                )
                attribute_data = attribute_data[3 + attribute_data_length :]
        elif command_id == CommandId.GET_APP_ATTRIBUTES:
            if 0 not in self._response_accumulator[1:]:
                # No null-terminated string yet.
                return

            app_identifier_length = self._response_accumulator.find(0, 1) - 1
            app_identifier = self._response_accumulator[
                1 : 1 + app_identifier_length
            ].decode("utf-8")
            if app_identifier != self._expected_response_app_identifier:
                logger.warning(
                    "unexpected response app identifier: "
                    f"expected {self._expected_response_app_identifier} "
                    f"but got {app_identifier}"
                )
                self._reset_response()
                if not self._response.done():
                    self._response.set_exception(ProtocolError())

            attribute_data = self._response_accumulator[1 + app_identifier_length + 1 :]
            while len(attribute_data) >= 3:
                attribute_id, attribute_data_length = struct.unpack_from(
                    "<BH", attribute_data, 0
                )
                if len(attribute_data) < 3 + attribute_data_length:
                    return
                attributes.append(
                    AppAttribute(
                        AppAttributeId(attribute_id),
                        attribute_data[3 : 3 + attribute_data_length].decode("utf-8"),
                    )
                )
                attribute_data = attribute_data[3 + attribute_data_length :]
        else:
            logger.warning(f"unexpected response command id {command_id}")
            return

        if len(attributes) < self._expected_response_tuples:
            # We have not received all the tuples yet.
            return

        if not self._response.done():
            self._response.set_result(attributes)

    async def _send_command(self, command: bytes) -> None:
        try:
            await self._ancs_proxy.control_point.write_value(
                command, with_response=True
            )
        except ATT_Error as error:
            raise CommandError(error_code=ErrorCode(error.error_code)) from error

    async def get_notification_attributes(
        self,
        notification_uid: int,
        attributes: Sequence[
            NotificationAttributeId | tuple[NotificationAttributeId, int]
        ],
    ) -> list[NotificationAttribute]:
        if not self._started:
            raise RuntimeError("client not started")

        command = struct.pack(
            "<BI", CommandId.GET_NOTIFICATION_ATTRIBUTES, notification_uid
        )
        for attribute in attributes:
            attribute_max_length = 0
            if isinstance(attribute, tuple):
                attribute_id, attribute_max_length = attribute
                if attribute_id not in (
                    NotificationAttributeId.TITLE,
                    NotificationAttributeId.SUBTITLE,
                    NotificationAttributeId.MESSAGE,
                ):
                    raise ValueError(
                        "this attribute does not allow specifying a max length"
                    )
            else:
                attribute_id = attribute
                if attribute_id in (
                    NotificationAttributeId.TITLE,
                    NotificationAttributeId.SUBTITLE,
                    NotificationAttributeId.MESSAGE,
                ):
                    attribute_max_length = _DEFAULT_ATTRIBUTE_MAX_LENGTH

            if attribute_max_length:
                command += struct.pack("<BH", attribute_id, attribute_max_length)
            else:
                command += struct.pack("B", attribute_id)

        try:
            async with self._command_semaphore:
                self._expected_response_notification_uid = notification_uid
                self._expected_response_tuples = len(attributes)
                self._expected_response_command_id = (
                    CommandId.GET_NOTIFICATION_ATTRIBUTES
                )
                self._response = asyncio.Future()

                # Send the command.
                await self._send_command(command)

                # Wait for the response.
                return await self._response
        finally:
            self._reset_response()

    async def get_app_attributes(
        self, app_identifier: str, attributes: Sequence[AppAttributeId]
    ) -> list[AppAttribute]:
        if not self._started:
            raise RuntimeError("client not started")

        command = (
            bytes([CommandId.GET_APP_ATTRIBUTES])
            + app_identifier.encode("utf-8")
            + b"\0"
        )
        for attribute_id in attributes:
            command += struct.pack("B", attribute_id)

        try:
            async with self._command_semaphore:
                self._expected_response_app_identifier = app_identifier
                self._expected_response_tuples = len(attributes)
                self._expected_response_command_id = CommandId.GET_APP_ATTRIBUTES
                self._response = asyncio.Future()

                # Send the command.
                await self._send_command(command)

                # Wait for the response.
                return await self._response
        finally:
            self._reset_response()

    async def perform_action(self, notification_uid: int, action: ActionId) -> None:
        if not self._started:
            raise RuntimeError("client not started")

        command = struct.pack(
            "<BIB", CommandId.PERFORM_NOTIFICATION_ACTION, notification_uid, action
        )

        async with self._command_semaphore:
            await self._send_command(command)

    async def perform_positive_action(self, notification_uid: int) -> None:
        return await self.perform_action(notification_uid, ActionId.POSITIVE)

    async def perform_negative_action(self, notification_uid: int) -> None:
        return await self.perform_action(notification_uid, ActionId.NEGATIVE)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: netsim/packet_streamer.proto
# Protobuf Python Version: 4.25.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from bumble.transport.grpc_protobuf.netsim import hci_packet_pb2 as netsim_dot_hci__packet__pb2
from bumble.transport.grpc_protobuf.netsim import startup_pb2 as netsim_dot_startup__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1cnetsim/packet_streamer.proto\x12\rnetsim.packet\x1a\x17netsim/hci_packet.proto\x1a\x14netsim/startup.proto\"\x93\x01\n\rPacketRequest\x12\x30\n\x0cinitial_info\x18\x01 \x01(\x0b\x32\x18.netsim.startup.ChipInfoH\x00\x12.\n\nhci_packet\x18\x02 \x01(\x0b\x32\x18.netsim.packet.HCIPacketH\x00\x12\x10\n\x06packet\x18\x03 \x01(\x0cH\x00\x42\x0e\n\x0crequest_type\"t\n\x0ePacketResponse\x12\x0f\n\x05\x65rror\x18\x01 \x01(\tH\x00\x12.\n\nhci_packet\x18\x02 \x01(\x0b\x32\x18.netsim.packet.HCIPacketH\x00\x12\x10\n\x06packet\x18\x03 \x01(\x0cH\x00\x42\x0f\n\rresponse_type2b\n\x0ePacketStreamer\x12P\n\rStreamPackets\x12\x1c.netsim.packet.PacketRequest\x1a\x1d.netsim.packet.PacketResponse(\x01\x30\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'netsim.packet_streamer_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_PACKETREQUEST']._serialized_start=95
  _globals['_PACKETREQUEST']._serialized_end=242
  _globals['_PACKETRESPONSE']._serialized_start=244
  _globals['_PACKETRESPONSE']._serialized_end=360
  _globals['_PACKETSTREAMER']._serialized_start=362
  _globals['_PACKETSTREAMER']._serialized_end=460
# @@protoc_insertion_point(module_scope)



# Copyright 2021-2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -----------------------------------------------------------------------------
# GATT - Generic Attribute Profile
# Client
#
# See Bluetooth spec @ Vol 3, Part G
#
# -----------------------------------------------------------------------------

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------
from __future__ import annotations

import asyncio
import functools
import logging
import struct
from collections.abc import Callable, Iterable
from datetime import datetime
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Generic,
    TypeVar,
    overload,
)

from typing_extensions import Self

from bumble import att, core, l2cap, utils
from bumble.colors import color
from bumble.core import UUID, InvalidStateError
from bumble.gatt import (
    GATT_CHARACTERISTIC_ATTRIBUTE_TYPE,
    GATT_CLIENT_CHARACTERISTIC_CONFIGURATION_DESCRIPTOR,
    GATT_INCLUDE_ATTRIBUTE_TYPE,
    GATT_PRIMARY_SERVICE_ATTRIBUTE_TYPE,
    GATT_REQUEST_TIMEOUT,
    GATT_SECONDARY_SERVICE_ATTRIBUTE_TYPE,
    Characteristic,
    ClientCharacteristicConfigurationBits,
    InvalidServiceError,
    TemplateService,
)
from bumble.hci import HCI_Constant

if TYPE_CHECKING:
    from bumble import device as device_module

# -----------------------------------------------------------------------------
# Typing
# -----------------------------------------------------------------------------
_T = TypeVar('_T')

# -----------------------------------------------------------------------------
# Logging
# -----------------------------------------------------------------------------
logger = logging.getLogger(__name__)


# -----------------------------------------------------------------------------
# Utils
# -----------------------------------------------------------------------------


def show_services(services: Iterable[ServiceProxy]) -> None:
    for service in services:
        print(color(str(service), 'cyan'))

        for characteristic in service.characteristics:
            print(color('  ' + str(characteristic), 'magenta'))

            for descriptor in characteristic.descriptors:
                print(color('    ' + str(descriptor), 'green'))


# -----------------------------------------------------------------------------
# Proxies
# -----------------------------------------------------------------------------
class AttributeProxy(utils.EventEmitter, Generic[_T]):
    def __init__(
        self, client: Client, handle: int, end_group_handle: int, attribute_type: UUID
    ) -> None:
        utils.EventEmitter.__init__(self)
        self.client = client
        self.handle = handle
        self.end_group_handle = end_group_handle
        self.type = attribute_type

    async def read_value(self, no_long_read: bool = False) -> _T:
        return self.decode_value(
            await self.client.read_value(self.handle, no_long_read)
        )

    async def write_value(self, value: _T, with_response=False):
        return await self.client.write_value(
            self.handle, self.encode_value(value), with_response
        )

    def encode_value(self, value: _T) -> bytes:
        return value  # type: ignore

    def decode_value(self, value: bytes) -> _T:
        return value  # type: ignore

    def __str__(self) -> str:
        return f'Attribute(handle=0x{self.handle:04X}, type={self.type})'


class ServiceProxy(AttributeProxy):
    uuid: UUID
    characteristics: list[CharacteristicProxy[bytes]]
    included_services: list[ServiceProxy]

    @staticmethod
    def from_client(service_class, client: Client, service_uuid: UUID):
        # The service and its characteristics are considered to have already been
        # discovered
        services = client.get_services_by_uuid(service_uuid)
        service = services[0] if services else None
        return service_class(service) if service else None

    def __init__(self, client, handle, end_group_handle, uuid, primary=True):
        attribute_type = (
            GATT_PRIMARY_SERVICE_ATTRIBUTE_TYPE
            if primary
            else GATT_SECONDARY_SERVICE_ATTRIBUTE_TYPE
        )
        super().__init__(client, handle, end_group_handle, attribute_type)
        self.uuid = uuid
        self.characteristics = []

    async def discover_characteristics(
        self, uuids=()
    ) -> list[CharacteristicProxy[bytes]]:
        return await self.client.discover_characteristics(uuids, self)

    def get_characteristics_by_uuid(
        self, uuid: UUID
    ) -> list[CharacteristicProxy[bytes]]:
        """Get all the characteristics with a specified UUID."""
        return self.client.get_characteristics_by_uuid(uuid, self)

    def get_required_characteristic_by_uuid(
        self, uuid: UUID
    ) -> CharacteristicProxy[bytes]:
        """
        Get the first characteristic with a specified UUID.

        If no characteristic with that UUID is found, an InvalidServiceError is raised.
        """
        if not (characteristics := self.get_characteristics_by_uuid(uuid)):
            raise InvalidServiceError(f'{uuid} characteristic not found')
        return characteristics[0]

    def __str__(self) -> str:
        return f'Service(handle=0x{self.handle:04X}, uuid={self.uuid})'


class CharacteristicProxy(AttributeProxy[_T]):
    properties: Characteristic.Properties
    descriptors: list[DescriptorProxy]
    subscribers: dict[Any, Callable[[_T], Any]]

    EVENT_UPDATE = "update"

    def __init__(
        self,
        client: Client,
        handle: int,
        end_group_handle: int,
        uuid: UUID,
        properties: int,
    ) -> None:
        super().__init__(client, handle, end_group_handle, uuid)
        self.uuid = uuid
        self.properties = Characteristic.Properties(properties)
        self.descriptors = []
        self.descriptors_discovered = False
        self.subscribers = {}  # Map from subscriber to proxy subscriber

    def get_descriptor(self, descriptor_type: UUID) -> DescriptorProxy | None:
        for descriptor in self.descriptors:
            if descriptor.type == descriptor_type:
                return descriptor

        return None

    async def discover_descriptors(self) -> list[DescriptorProxy]:
        return await self.client.discover_descriptors(self)

    async def subscribe(
        self,
        subscriber: Callable[[_T], Any] | None = None,
        prefer_notify: bool = True,
    ) -> None:
        if subscriber is not None:
            if subscriber in self.subscribers:
                # We already have a proxy subscriber
                subscriber = self.subscribers[subscriber]
            else:
                # Create and register a proxy that will decode the value
                original_subscriber = subscriber

                def on_change(value):
                    original_subscriber(self.decode_value(value))

                self.subscribers[subscriber] = on_change
                subscriber = on_change

        await self.client.subscribe(self, subscriber, prefer_notify)

    async def unsubscribe(self, subscriber=None, force=False) -> None:
        if subscriber in self.subscribers:
            subscriber = self.subscribers.pop(subscriber)

        await self.client.unsubscribe(self, subscriber, force)

    def __str__(self) -> str:
        return (
            f'Characteristic(handle=0x{self.handle:04X}, '
            f'uuid={self.uuid}, '
            f'{self.properties!s})'
        )


class DescriptorProxy(AttributeProxy[bytes]):
    def __init__(self, client: Client, handle: int, descriptor_type: UUID) -> None:
        super().__init__(client, handle, 0, descriptor_type)

    def __str__(self) -> str:
        return f'Descriptor(handle=0x{self.handle:04X}, type={self.type})'


class ProfileServiceProxy:
    '''
    Base class for profile-specific service proxies
    '''

    SERVICE_CLASS: ClassVar[type[TemplateService]]

    @classmethod
    def from_client(cls, client: Client) -> Self | None:
        return ServiceProxy.from_client(cls, client, cls.SERVICE_CLASS.UUID)


# -----------------------------------------------------------------------------
# GATT Client
# -----------------------------------------------------------------------------
class Client:
    services: list[ServiceProxy]
    cached_values: dict[int, tuple[datetime, bytes]]
    notification_subscribers: dict[
        int, set[CharacteristicProxy | Callable[[bytes], Any]]
    ]
    indication_subscribers: dict[int, set[CharacteristicProxy | Callable[[bytes], Any]]]
    pending_response: asyncio.futures.Future[att.ATT_PDU] | None
    pending_request: att.ATT_PDU | None

    def __init__(self, bearer: att.Bearer) -> None:
        self.bearer = bearer
        self.mtu_exchange_done = False
        self.request_semaphore = asyncio.Semaphore(1)
        self.pending_request = None
        self.pending_response = None
        self.notification_subscribers = {}  # Subscriber set, by attribute handle
        self.indication_subscribers = {}  # Subscriber set, by attribute handle
        self.services = []
        self.cached_values = {}

        if att.is_enhanced_bearer(bearer):
            bearer.on(bearer.EVENT_CLOSE, self.on_disconnection)
            self._bearer_id = (
                f'[0x{bearer.connection.handle:04X}|CID=0x{bearer.source_cid:04X}]'
            )
            self.connection = bearer.connection
        else:
            bearer.on(bearer.EVENT_DISCONNECTION, self.on_disconnection)
            self._bearer_id = f'[0x{bearer.handle:04X}]'
            self.connection = bearer

    @overload
    @classmethod
    async def connect_eatt(
        cls,
        connection: device_module.Connection,
        spec: l2cap.LeCreditBasedChannelSpec | None = None,
    ) -> Client: ...

    @overload
    @classmethod
    async def connect_eatt(
        cls,
        connection: device_module.Connection,
        spec: l2cap.LeCreditBasedChannelSpec | None = None,
        count: int = 1,
    ) -> list[Client]: ...

    @classmethod
    async def connect_eatt(
        cls,
        connection: device_module.Connection,
        spec: l2cap.LeCreditBasedChannelSpec | None = None,
        count: int = 1,
    ) -> list[Client] | Client:
        channels = await connection.device.l2cap_channel_manager.create_enhanced_credit_based_channels(
            connection,
            spec or l2cap.LeCreditBasedChannelSpec(psm=att.EATT_PSM),
            count,
        )

        def on_pdu(client: Client, pdu: bytes):
            client.on_gatt_pdu(att.ATT_PDU.from_bytes(pdu))

        clients = [cls(channel) for channel in channels]
        for channel, client in zip(channels, clients):
            channel.sink = functools.partial(on_pdu, client)
            channel.att_mtu = att.ATT_DEFAULT_MTU
        return clients[0] if count == 1 else clients

    @property
    def mtu(self) -> int:
        return self.bearer.att_mtu

    @mtu.setter
    def mtu(self, value: int) -> None:
        self.bearer.on_att_mtu_update(value)

    def send_gatt_pdu(self, pdu: bytes) -> None:
        if att.is_enhanced_bearer(self.bearer):
            self.bearer.write(pdu)
        else:
            self.bearer.send_l2cap_pdu(att.ATT_CID, pdu)

    async def send_command(self, command: att.ATT_PDU) -> None:
        logger.debug(f'GATT Command from client: {self._bearer_id} {command}')
        self.send_gatt_pdu(bytes(command))

    async def send_request(self, request: att.ATT_PDU):
        logger.debug(f'GATT Request from client: {self._bearer_id} {request}')

        # Wait until we can send (only one pending command at a time for the connection)
        response = None
        async with self.request_semaphore:
            assert self.pending_request is None
            assert self.pending_response is None

            # Create a future value to hold the eventual response
            self.pending_response = asyncio.get_running_loop().create_future()
            self.pending_request = request

            try:
                self.send_gatt_pdu(bytes(request))
                response = await asyncio.wait_for(
                    self.pending_response, GATT_REQUEST_TIMEOUT
                )
            except asyncio.TimeoutError as error:
                logger.warning(color('!!! GATT Request timeout', 'red'))
                raise core.TimeoutError(f'GATT timeout for {request.name}') from error
            finally:
                self.pending_request = None
                self.pending_response = None

        return response

    def send_confirmation(
        self, confirmation: att.ATT_Handle_Value_Confirmation
    ) -> None:
        logger.debug(f'GATT Confirmation from client: {self._bearer_id} {confirmation}')
        self.send_gatt_pdu(bytes(confirmation))

    async def request_mtu(self, mtu: int) -> int:
        # Check the range
        if mtu < att.ATT_DEFAULT_MTU:
            raise core.InvalidArgumentError(f'MTU must be >= {att.ATT_DEFAULT_MTU}')
        if mtu > 0xFFFF:
            raise core.InvalidArgumentError('MTU must be <= 0xFFFF')

        # We can only send one request per connection
        if self.mtu_exchange_done:
            return self.mtu

        # Send the request
        self.mtu_exchange_done = True
        response = await self.send_request(
            att.ATT_Exchange_MTU_Request(client_rx_mtu=mtu)
        )
        if response.op_code == att.Opcode.ATT_ERROR_RESPONSE:
            raise att.ATT_Error(error_code=response.error_code, message=response)

        # Compute the final MTU
        self.mtu = min(mtu, response.server_rx_mtu)

        return self.mtu

    def get_services_by_uuid(self, uuid: UUID) -> list[ServiceProxy]:
        return [service for service in self.services if service.uuid == uuid]

    def get_characteristics_by_uuid(
        self, uuid: UUID, service: ServiceProxy | None = None
    ) -> list[CharacteristicProxy[bytes]]:
        services = [service] if service else self.services
        return [
            c
            for c in [c for s in services for c in s.characteristics]
            if c.uuid == uuid
        ]

    def get_attribute_grouping(
        self, attribute_handle: int
    ) -> (
        ServiceProxy
        | tuple[ServiceProxy, CharacteristicProxy]
        | tuple[ServiceProxy, CharacteristicProxy, DescriptorProxy]
        | None
    ):
        """
        Get the attribute(s) associated with an attribute handle
        """
        for service in self.services:
            if service.handle == attribute_handle:
                return service
            if service.handle <= attribute_handle <= service.end_group_handle:
                for characteristic in service.characteristics:
                    if characteristic.handle == attribute_handle:
                        return (service, characteristic)
                    if (
                        characteristic.handle
                        <= attribute_handle
                        <= characteristic.end_group_handle
                    ):
                        for descriptor in characteristic.descriptors:
                            if descriptor.handle == attribute_handle:
                                return (service, characteristic, descriptor)
        return None

    def on_service_discovered(self, service):
        '''Add a service to the service list if it wasn't already there'''
        already_known = False
        for existing_service in self.services:
            if existing_service.handle == service.handle:
                already_known = True
                break
        if not already_known:
            self.services.append(service)

    async def discover_services(self, uuids: Iterable[UUID] = ()) -> list[ServiceProxy]:
        '''
        See Vol 3, Part G - 4.4.1 Discover All Primary Services
        '''
        starting_handle = 0x0001
        services = []
        while starting_handle < 0xFFFF:
            response = await self.send_request(
                att.ATT_Read_By_Group_Type_Request(
                    starting_handle=starting_handle,
                    ending_handle=0xFFFF,
                    attribute_group_type=GATT_PRIMARY_SERVICE_ATTRIBUTE_TYPE,
                )
            )
            if response is None:
                # TODO raise appropriate exception
                return []

            # Check if we reached the end of the iteration
            if response.op_code == att.Opcode.ATT_ERROR_RESPONSE:
                if response.error_code != att.ATT_ATTRIBUTE_NOT_FOUND_ERROR:
                    # Unexpected end
                    logger.warning(
                        '!!! unexpected error while discovering services: '
                        f'{HCI_Constant.error_name(response.error_code)}'
                    )
                    raise att.ATT_Error(
                        error_code=response.error_code,
                        message='Unexpected error while discovering services',
                    )
                break

            for (
                attribute_handle,
                end_group_handle,
                attribute_value,
            ) in response.attributes:
                if (
                    attribute_handle < starting_handle
                    or end_group_handle < attribute_handle
                ):
                    # Something's not right
                    logger.warning(
                        f'bogus handle values: {attribute_handle} {end_group_handle}'
                    )
                    return []

                # Create a service proxy for this service
                service = ServiceProxy(
                    self,
                    attribute_handle,
                    end_group_handle,
                    UUID.from_bytes(attribute_value),
                    True,
                )

                # Filter out returned services based on the given uuids list
                if (not uuids) or (service.uuid in uuids):
                    services.append(service)

                # Add the service to the peer's service list
                self.on_service_discovered(service)

            # Stop if for some reason the list was empty
            if not response.attributes:
                break

            # Move on to the next chunk
            starting_handle = response.attributes[-1][1] + 1

        return services

    async def discover_service(self, uuid: str | UUID) -> list[ServiceProxy]:
        '''
        See Vol 3, Part G - 4.4.2 Discover Primary Service by Service UUID
        '''

        # Force uuid to be a UUID object
        if isinstance(uuid, str):
            uuid = UUID(uuid)

        starting_handle = 0x0001
        services = []
        while starting_handle < 0xFFFF:
            response = await self.send_request(
                att.ATT_Find_By_Type_Value_Request(
                    starting_handle=starting_handle,
                    ending_handle=0xFFFF,
                    attribute_type=GATT_PRIMARY_SERVICE_ATTRIBUTE_TYPE,
                    attribute_value=uuid.to_pdu_bytes(),
                )
            )
            if response is None:
                # TODO raise appropriate exception
                return []

            # Check if we reached the end of the iteration
            if response.op_code == att.Opcode.ATT_ERROR_RESPONSE:
                if response.error_code != att.ATT_ATTRIBUTE_NOT_FOUND_ERROR:
                    # Unexpected end
                    logger.warning(
                        '!!! unexpected error while discovering services: '
                        f'{HCI_Constant.error_name(response.error_code)}'
                    )
                    # TODO raise appropriate exception
                    return []
                break

            for attribute_handle, end_group_handle in response.handles_information:
                if (
                    attribute_handle < starting_handle
                    or end_group_handle < attribute_handle
                ):
                    # Something's not right
                    logger.warning(
                        f'bogus handle values: {attribute_handle} {end_group_handle}'
                    )
                    return []

                # Create a service proxy for this service
                service = ServiceProxy(
                    self, attribute_handle, end_group_handle, uuid, True
                )

                # Add the service to the peer's service list
                services.append(service)
                self.on_service_discovered(service)

                # Check if we've reached the end already
                if end_group_handle == 0xFFFF:
                    break

            # Stop if for some reason the list was empty
            if not response.handles_information:
                break

            # Move on to the next chunk
            starting_handle = response.handles_information[-1][1] + 1

        return services

    async def discover_included_services(
        self, service: ServiceProxy
    ) -> list[ServiceProxy]:
        '''
        See Vol 3, Part G - 4.5.1 Find Included Services
        '''

        starting_handle = service.handle
        ending_handle = service.end_group_handle

        included_services: list[ServiceProxy] = []
        while starting_handle <= ending_handle:
            response = await self.send_request(
                att.ATT_Read_By_Type_Request(
                    starting_handle=starting_handle,
                    ending_handle=ending_handle,
                    attribute_type=GATT_INCLUDE_ATTRIBUTE_TYPE,
                )
            )
            if response is None:
                # TODO raise appropriate exception
                return []

            # Check if we reached the end of the iteration
            if response.op_code == att.Opcode.ATT_ERROR_RESPONSE:
                if response.error_code != att.ATT_ATTRIBUTE_NOT_FOUND_ERROR:
                    # Unexpected end
                    logger.warning(
                        '!!! unexpected error while discovering included services: '
                        f'{HCI_Constant.error_name(response.error_code)}'
                    )
                    raise att.ATT_Error(
                        error_code=response.error_code,
                        message='Unexpected error while discovering included services',
                    )
                break

            # Stop if for some reason the list was empty
            if not response.attributes:
                break

            # Process all included services returned in this iteration
            for attribute_handle, attribute_value in response.attributes:
                if attribute_handle < starting_handle:
                    # Something's not right
                    logger.warning(f'bogus handle value: {attribute_handle}')
                    return []

                group_starting_handle, group_ending_handle = struct.unpack_from(
                    '<HH', attribute_value
                )
                service_uuid = UUID.from_bytes(attribute_value[4:])
                included_service = ServiceProxy(
                    self, group_starting_handle, group_ending_handle, service_uuid, True
                )

                included_services.append(included_service)

            # Move on to the next included services
            starting_handle = response.attributes[-1][0] + 1

        service.included_services = included_services
        return included_services

    async def discover_characteristics(
        self, uuids, service: ServiceProxy | None
    ) -> list[CharacteristicProxy[bytes]]:
        '''
        See Vol 3, Part G - 4.6.1 Discover All Characteristics of a Service and 4.6.2
        Discover Characteristics by UUID
        '''

        # Cast the UUIDs type from string to object if needed
        uuids = [UUID(uuid) if isinstance(uuid, str) else uuid for uuid in uuids]

        # Decide which services to discover for
        services = [service] if service else self.services

        # Perform characteristic discovery for each service
        discovered_characteristics: list[CharacteristicProxy[bytes]] = []
        for service in services:
            starting_handle = service.handle
            ending_handle = service.end_group_handle

            characteristics: list[CharacteristicProxy[bytes]] = []
            while starting_handle <= ending_handle:
                response = await self.send_request(
                    att.ATT_Read_By_Type_Request(
                        starting_handle=starting_handle,
                        ending_handle=ending_handle,
                        attribute_type=GATT_CHARACTERISTIC_ATTRIBUTE_TYPE,
                    )
                )
                if response is None:
                    # TODO raise appropriate exception
                    return []

                # Check if we reached the end of the iteration
                if response.op_code == att.Opcode.ATT_ERROR_RESPONSE:
                    if response.error_code != att.ATT_ATTRIBUTE_NOT_FOUND_ERROR:
                        # Unexpected end
                        logger.warning(
                            '!!! unexpected error while discovering characteristics: '
                            f'{HCI_Constant.error_name(response.error_code)}'
                        )
                        raise att.ATT_Error(
                            error_code=response.error_code,
                            message='Unexpected error while discovering characteristics',
                        )
                    break

                # Stop if for some reason the list was empty
                if not response.attributes:
                    break

                # Process all characteristics returned in this iteration
                for attribute_handle, attribute_value in response.attributes:
                    if attribute_handle < starting_handle:
                        # Something's not right
                        logger.warning(f'bogus handle value: {attribute_handle}')
                        return []

                    properties, handle = struct.unpack_from('<BH', attribute_value)
                    characteristic_uuid = UUID.from_bytes(attribute_value[3:])
                    characteristic = CharacteristicProxy[bytes](
                        self, handle, 0, characteristic_uuid, properties
                    )

                    # Set the previous characteristic's end handle
                    if characteristics:
                        characteristics[-1].end_group_handle = attribute_handle - 1

                    characteristics.append(characteristic)

                # Move on to the next characteristics
                starting_handle = response.attributes[-1][0] + 1

            # Set the end handle for the last characteristic
            if characteristics:
                characteristics[-1].end_group_handle = service.end_group_handle

            # Set the service's characteristics
            characteristics = [
                c for c in characteristics if not uuids or c.uuid in uuids
            ]
            service.characteristics = characteristics
            discovered_characteristics.extend(characteristics)

        return discovered_characteristics

    async def discover_descriptors(
        self,
        characteristic: CharacteristicProxy | None = None,
        start_handle: int | None = None,
        end_handle: int | None = None,
    ) -> list[DescriptorProxy]:
        '''
        See Vol 3, Part G - 4.7.1 Discover All Characteristic Descriptors
        '''
        if characteristic:
            starting_handle = characteristic.handle + 1
            ending_handle = characteristic.end_group_handle
        elif start_handle and end_handle:
            starting_handle = start_handle
            ending_handle = end_handle
        else:
            return []

        descriptors: list[DescriptorProxy] = []
        while starting_handle <= ending_handle:
            response = await self.send_request(
                att.ATT_Find_Information_Request(
                    starting_handle=starting_handle, ending_handle=ending_handle
                )
            )
            if response is None:
                # TODO raise appropriate exception
                return []

            # Check if we reached the end of the iteration
            if response.op_code == att.Opcode.ATT_ERROR_RESPONSE:
                if response.error_code != att.ATT_ATTRIBUTE_NOT_FOUND_ERROR:
                    # Unexpected end
                    logger.warning(
                        '!!! unexpected error while discovering descriptors: '
                        f'{HCI_Constant.error_name(response.error_code)}'
                    )
                    # TODO raise appropriate exception
                    return []
                break

            # Stop if for some reason the list was empty
            if not response.information:
                break

            # Process all descriptors returned in this iteration
            for attribute_handle, attribute_uuid in response.information:
                if attribute_handle < starting_handle:
                    # Something's not right
                    logger.warning(f'bogus handle value: {attribute_handle}')
                    return []

                descriptor = DescriptorProxy(
                    self, attribute_handle, UUID.from_bytes(attribute_uuid)
                )
                descriptors.append(descriptor)
                # TODO: read descriptor value

            # Move on to the next descriptor
            starting_handle = response.information[-1][0] + 1

        # Set the characteristic's descriptors
        if characteristic:
            characteristic.descriptors = descriptors

        return descriptors

    async def discover_attributes(self) -> list[AttributeProxy[bytes]]:
        '''
        Discover all attributes, regardless of type
        '''
        starting_handle = 0x0001
        ending_handle = 0xFFFF
        attributes = []
        while True:
            response = await self.send_request(
                att.ATT_Find_Information_Request(
                    starting_handle=starting_handle, ending_handle=ending_handle
                )
            )
            if response is None:
                return []

            # Check if we reached the end of the iteration
            if response.op_code == att.Opcode.ATT_ERROR_RESPONSE:
                if response.error_code != att.ATT_ATTRIBUTE_NOT_FOUND_ERROR:
                    # Unexpected end
                    logger.warning(
                        '!!! unexpected error while discovering attributes: '
                        f'{HCI_Constant.error_name(response.error_code)}'
                    )
                    return []
                break

            for attribute_handle, attribute_uuid in response.information:
                if attribute_handle < starting_handle:
                    # Something's not right
                    logger.warning(f'bogus handle value: {attribute_handle}')
                    return []

                attribute = AttributeProxy[bytes](
                    self, attribute_handle, 0, UUID.from_bytes(attribute_uuid)
                )
                attributes.append(attribute)

            # Move on to the next attributes
            starting_handle = attributes[-1].handle + 1

        return attributes

    async def subscribe(
        self,
        characteristic: CharacteristicProxy,
        subscriber: Callable[[Any], Any] | None = None,
        prefer_notify: bool = True,
    ) -> None:
        # If we haven't already discovered the descriptors for this characteristic,
        # do it now
        if not characteristic.descriptors_discovered:
            await self.discover_descriptors(characteristic)

        # Look for the CCCD descriptor
        cccd = characteristic.get_descriptor(
            GATT_CLIENT_CHARACTERISTIC_CONFIGURATION_DESCRIPTOR
        )
        if not cccd:
            logger.warning('subscribing to characteristic with no CCCD descriptor')
            return

        if (
            characteristic.properties & Characteristic.Properties.NOTIFY
            and characteristic.properties & Characteristic.Properties.INDICATE
        ):
            if prefer_notify:
                bits = ClientCharacteristicConfigurationBits.NOTIFICATION
                subscribers = self.notification_subscribers
            else:
                bits = ClientCharacteristicConfigurationBits.INDICATION
                subscribers = self.indication_subscribers
        elif characteristic.properties & Characteristic.Properties.NOTIFY:
            bits = ClientCharacteristicConfigurationBits.NOTIFICATION
            subscribers = self.notification_subscribers
        elif characteristic.properties & Characteristic.Properties.INDICATE:
            bits = ClientCharacteristicConfigurationBits.INDICATION
            subscribers = self.indication_subscribers
        else:
            raise InvalidStateError("characteristic is not notify or indicate")

        # Add subscribers to the sets
        subscriber_set = subscribers.setdefault(characteristic.handle, set())
        if subscriber is not None:
            subscriber_set.add(subscriber)

        # Add the characteristic as a subscriber, which will result in the
        # characteristic emitting an 'update' event when a notification or indication
        # is received
        subscriber_set.add(characteristic)

        await self.write_value(cccd, struct.pack('<H', bits), with_response=True)

    async def unsubscribe(
        self,
        characteristic: CharacteristicProxy,
        subscriber: Callable[[Any], Any] | None = None,
        force: bool = False,
    ) -> None:
        '''
        Unsubscribe from a characteristic.

        If `force` is True, this will write zeros to the CCCD when there are no
        subscribers left, even if there were already no registered subscribers.
        '''
        # If we haven't already discovered the descriptors for this characteristic,
        # do it now
        if not characteristic.descriptors_discovered:
            await self.discover_descriptors(characteristic)

        # Look for the CCCD descriptor
        cccd = characteristic.get_descriptor(
            GATT_CLIENT_CHARACTERISTIC_CONFIGURATION_DESCRIPTOR
        )
        if not cccd:
            logger.warning('unsubscribing from characteristic with no CCCD descriptor')
            return

        # Check if the characteristic has subscribers
        if not (
            characteristic.handle in self.notification_subscribers
            or characteristic.handle in self.indication_subscribers
        ):
            if not force:
                return

        # Remove the subscriber(s)
        if subscriber is not None:
            # Remove matching subscriber from subscriber sets
            for subscriber_set in (
                self.notification_subscribers,
                self.indication_subscribers,
            ):
                if (
                    subscribers := subscriber_set.get(characteristic.handle)
                ) and subscriber in subscribers:
                    subscribers.remove(subscriber)

                    # The characteristic itself is added as subscriber. If it is the
                    # last remaining subscriber, we remove it, such that the clean up
                    # works correctly. Otherwise the CCCD never is set back to 0.
                    if len(subscribers) == 1 and characteristic in subscribers:
                        subscribers.remove(characteristic)

                    # Cleanup if we removed the last one
                    if not subscribers:
                        del subscriber_set[characteristic.handle]
        else:
            # Remove all subscribers for this attribute from the sets
            self.notification_subscribers.pop(characteristic.handle, None)
            self.indication_subscribers.pop(characteristic.handle, None)

        # Update the CCCD
        if not (
            characteristic.handle in self.notification_subscribers
            or characteristic.handle in self.indication_subscribers
        ):
            # No more subscribers left
            await self.write_value(cccd, b'\x00\x00', with_response=True)

    async def read_value(
        self, attribute: int | AttributeProxy, no_long_read: bool = False
    ) -> bytes:
        '''
        See Vol 3, Part G - 4.8.1 Read Characteristic Value

        `attribute` can be an Attribute object, or a handle value
        '''

        # Send a request to read
        attribute_handle = attribute if isinstance(attribute, int) else attribute.handle
        response = await self.send_request(
            att.ATT_Read_Request(attribute_handle=attribute_handle)
        )
        if response is None:
            raise TimeoutError('read timeout')
        if response.op_code == att.Opcode.ATT_ERROR_RESPONSE:
            raise att.ATT_Error(error_code=response.error_code, message=response)

        # If the value is the max size for the MTU, try to read more unless the caller
        # specifically asked not to do that
        attribute_value = response.attribute_value
        if not no_long_read and len(attribute_value) == self.mtu - 1:
            logger.debug('using READ BLOB to get the rest of the value')
            offset = len(attribute_value)
            while True:
                response = await self.send_request(
                    att.ATT_Read_Blob_Request(
                        attribute_handle=attribute_handle, value_offset=offset
                    )
                )
                if response is None:
                    raise TimeoutError('read timeout')
                if response.op_code == att.Opcode.ATT_ERROR_RESPONSE:
                    if response.error_code in (
                        att.ATT_ATTRIBUTE_NOT_LONG_ERROR,
                        att.ATT_INVALID_OFFSET_ERROR,
                    ):
                        break
                    raise att.ATT_Error(
                        error_code=response.error_code, message=response
                    )

                part = response.part_attribute_value
                attribute_value += part

                if len(part) < self.mtu - 1:
                    break

                offset += len(part)

        self.cache_value(attribute_handle, attribute_value)
        # Return the value as bytes
        return attribute_value

    async def read_characteristics_by_uuid(
        self, uuid: UUID, service: ServiceProxy | None
    ) -> list[bytes]:
        '''
        See Vol 3, Part G - 4.8.2 Read Using Characteristic UUID
        '''

        if service is None:
            starting_handle = 0x0001
            ending_handle = 0xFFFF
        else:
            starting_handle = service.handle
            ending_handle = service.end_group_handle

        characteristics_values = []
        while starting_handle <= ending_handle:
            response = await self.send_request(
                att.ATT_Read_By_Type_Request(
                    starting_handle=starting_handle,
                    ending_handle=ending_handle,
                    attribute_type=uuid,
                )
            )
            if response is None:
                # TODO raise appropriate exception
                return []

            # Check if we reached the end of the iteration
            if response.op_code == att.Opcode.ATT_ERROR_RESPONSE:
                if response.error_code != att.ATT_ATTRIBUTE_NOT_FOUND_ERROR:
                    # Unexpected end
                    logger.warning(
                        '!!! unexpected error while reading characteristics: '
                        f'{HCI_Constant.error_name(response.error_code)}'
                    )
                    # TODO raise appropriate exception
                    return []
                break

            # Stop if for some reason the list was empty
            if not response.attributes:
                break

            # Process all characteristics returned in this iteration
            for attribute_handle, attribute_value in response.attributes:
                if attribute_handle < starting_handle:
                    # Something's not right
                    logger.warning(f'bogus handle value: {attribute_handle}')
                    return []

                characteristics_values.append(attribute_value)

            # Move on to the next characteristics
            starting_handle = response.attributes[-1][0] + 1

        return characteristics_values

    async def write_value(
        self,
        attribute: int | AttributeProxy,
        value: bytes,
        with_response: bool = False,
    ) -> None:
        '''
        See Vol 3, Part G - 4.9.1 Write Without Response & 4.9.3 Write Characteristic
        Value

        `attribute` can be an Attribute object, or a handle value
        '''

        # Send a request or command to write
        attribute_handle = attribute if isinstance(attribute, int) else attribute.handle
        if with_response:
            response = await self.send_request(
                att.ATT_Write_Request(
                    attribute_handle=attribute_handle, attribute_value=value
                )
            )
            if response.op_code == att.Opcode.ATT_ERROR_RESPONSE:
                raise att.ATT_Error(error_code=response.error_code, message=response)
        else:
            await self.send_command(
                att.ATT_Write_Command(
                    attribute_handle=attribute_handle, attribute_value=value
                )
            )

    def on_disconnection(self, *args) -> None:
        del args  # unused.
        if self.pending_response and not self.pending_response.done():
            self.pending_response.cancel()

    def on_gatt_pdu(self, att_pdu: att.ATT_PDU) -> None:
        logger.debug(f'GATT Response to client: {self._bearer_id} {att_pdu}')
        if att_pdu.op_code in att.ATT_RESPONSES:
            if self.pending_request is None:
                # Not expected!
                logger.warning('!!! unexpected response, there is no pending request')
                return

            # The response should match the pending request unless it is
            # an error response
            if att_pdu.op_code != att.Opcode.ATT_ERROR_RESPONSE:
                expected_response_name = self.pending_request.name.replace(
                    '_REQUEST', '_RESPONSE'
                )
                if att_pdu.name != expected_response_name:
                    logger.warning(
                        f'!!! mismatched response: expected {expected_response_name}'
                    )
                    return

            # Return the response to the coroutine that is waiting for it
            assert self.pending_response is not None
            self.pending_response.set_result(att_pdu)
        else:
            handler_name = f'on_{att_pdu.name.lower()}'
            handler = getattr(self, handler_name, None)
            if handler is not None:
                handler(att_pdu)
            else:
                logger.warning(
                    color(
                        '--- Ignoring GATT Response from ' f'{self._bearer_id}: ',
                        'red',
                    )
                    + str(att_pdu)
                )

    def on_att_handle_value_notification(
        self, notification: att.ATT_Handle_Value_Notification
    ):
        # Call all subscribers
        subscribers = self.notification_subscribers.get(
            notification.attribute_handle, set()
        )
        if not subscribers:
            logger.warning('!!! received notification with no subscriber')

        self.cache_value(notification.attribute_handle, notification.attribute_value)
        for subscriber in subscribers:
            if callable(subscriber):
                subscriber(notification.attribute_value)
            else:
                subscriber.emit(subscriber.EVENT_UPDATE, notification.attribute_value)

    def on_att_handle_value_indication(
        self, indication: att.ATT_Handle_Value_Indication
    ):
        # Call all subscribers
        subscribers = self.indication_subscribers.get(
            indication.attribute_handle, set()
        )
        if not subscribers:
            logger.warning('!!! received indication with no subscriber')

        self.cache_value(indication.attribute_handle, indication.attribute_value)
        for subscriber in subscribers:
            if callable(subscriber):
                subscriber(indication.attribute_value)
            else:
                subscriber.emit(subscriber.EVENT_UPDATE, indication.attribute_value)

        # Confirm that we received the indication
        self.send_confirmation(att.ATT_Handle_Value_Confirmation())

    def cache_value(self, attribute_handle: int, value: bytes) -> None:
        self.cached_values[attribute_handle] = (
            datetime.now(),
            value,
        )



# Copyright 2021-2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------
import asyncio
import sys

import bumble.logging
from bumble.a2dp import make_audio_source_service_sdp_records
from bumble.avdtp import Protocol as AVDTP_Protocol
from bumble.colors import color
from bumble.core import (
    BT_AUDIO_SINK_SERVICE,
    BT_AVDTP_PROTOCOL_ID,
    BT_L2CAP_PROTOCOL_ID,
    PhysicalTransport,
)
from bumble.device import Device
from bumble.sdp import (
    SDP_BLUETOOTH_PROFILE_DESCRIPTOR_LIST_ATTRIBUTE_ID,
    SDP_PROTOCOL_DESCRIPTOR_LIST_ATTRIBUTE_ID,
    SDP_SERVICE_CLASS_ID_LIST_ATTRIBUTE_ID,
    DataElement,
    ServiceAttribute,
)
from bumble.sdp import Client as SDP_Client
from bumble.transport import open_transport


# -----------------------------------------------------------------------------
def sdp_records():
    service_record_handle = 0x00010001
    return {
        service_record_handle: make_audio_source_service_sdp_records(
            service_record_handle
        )
    }


# -----------------------------------------------------------------------------
# pylint: disable-next=too-many-nested-blocks
async def find_a2dp_service(connection):
    # Connect to the SDP Server
    sdp_client = SDP_Client(connection)
    await sdp_client.connect()

    # Search for services with an Audio Sink service class
    search_result = await sdp_client.search_attributes(
        [BT_AUDIO_SINK_SERVICE],
        [
            SDP_PROTOCOL_DESCRIPTOR_LIST_ATTRIBUTE_ID,
            SDP_BLUETOOTH_PROFILE_DESCRIPTOR_LIST_ATTRIBUTE_ID,
            SDP_SERVICE_CLASS_ID_LIST_ATTRIBUTE_ID,
        ],
    )

    print(color('==================================', 'blue'))
    print(color('A2DP Sink Services:', 'yellow'))

    service_version = None

    for attribute_list in search_result:
        print(color('SERVICE:', 'green'))

        # Service classes
        service_class_id_list = ServiceAttribute.find_attribute_in_list(
            attribute_list, SDP_SERVICE_CLASS_ID_LIST_ATTRIBUTE_ID
        )
        if service_class_id_list:
            if service_class_id_list.value:
                print(color('  Service Classes:', 'green'))
                for service_class_id in service_class_id_list.value:
                    print('   ', service_class_id.value)

        # Protocol info
        protocol_descriptor_list = ServiceAttribute.find_attribute_in_list(
            attribute_list, SDP_PROTOCOL_DESCRIPTOR_LIST_ATTRIBUTE_ID
        )
        if protocol_descriptor_list:
            print(color('  Protocol:', 'green'))
            for protocol_descriptor in protocol_descriptor_list.value:
                if protocol_descriptor.value[0].value == BT_L2CAP_PROTOCOL_ID:
                    if len(protocol_descriptor.value) >= 2:
                        psm = protocol_descriptor.value[1].value
                        print(f'{color("    L2CAP PSM:", "cyan")}     {psm}')
                elif protocol_descriptor.value[0].value == BT_AVDTP_PROTOCOL_ID:
                    if len(protocol_descriptor.value) >= 2:
                        avdtp_version_major = protocol_descriptor.value[1].value >> 8
                        avdtp_version_minor = protocol_descriptor.value[1].value & 0xFF
                        print(
                            f'{color("    AVDTP Version:", "cyan")} '
                            f'{avdtp_version_major}.{avdtp_version_minor}'
                        )
                        service_version = (avdtp_version_major, avdtp_version_minor)

        # Profile info
        bluetooth_profile_descriptor_list = ServiceAttribute.find_attribute_in_list(
            attribute_list, SDP_BLUETOOTH_PROFILE_DESCRIPTOR_LIST_ATTRIBUTE_ID
        )
        if bluetooth_profile_descriptor_list:
            if bluetooth_profile_descriptor_list.value:
                if (
                    bluetooth_profile_descriptor_list.value[0].type
                    == DataElement.SEQUENCE
                ):
                    bluetooth_profile_descriptors = (
                        bluetooth_profile_descriptor_list.value
                    )
                else:
                    # Sometimes, instead of a list of lists, we just find a list.
                    # Fix that.
                    bluetooth_profile_descriptors = [bluetooth_profile_descriptor_list]

                print(color('  Profiles:', 'green'))
                for bluetooth_profile_descriptor in bluetooth_profile_descriptors:
                    version_major = bluetooth_profile_descriptor.value[1].value >> 8
                    version_minor = bluetooth_profile_descriptor.value[1].value & 0xFF
                    print(
                        f'    {bluetooth_profile_descriptor.value[0].value}'
                        f' - version {version_major}.{version_minor}'
                    )

    await sdp_client.disconnect()
    return service_version


# -----------------------------------------------------------------------------
async def main() -> None:
    if len(sys.argv) < 4:
        print('Usage: run_a2dp_info.py <device-config> <transport-spec> <bt-addr>')
        print('example: run_a2dp_info.py classic1.json usb:0 14:7D:DA:4E:53:A8')
        return

    print('<<< connecting to HCI...')
    async with await open_transport(sys.argv[2]) as hci_transport:
        print('<<< connected')

        # Create a device
        device = Device.from_config_file_with_hci(
            sys.argv[1], hci_transport.source, hci_transport.sink
        )
        device.classic_enabled = True

        # Start the controller
        await device.power_on()

        # Setup the SDP to expose a SRC service, in case the remote device queries us
        # back
        device.sdp_service_records = sdp_records()

        # Connect to a peer
        target_address = sys.argv[3]
        print(f'=== Connecting to {target_address}...')
        connection = await device.connect(
            target_address, transport=PhysicalTransport.BR_EDR
        )
        print(f'=== Connected to {connection.peer_address}!')

        # Request authentication
        print('*** Authenticating...')
        await connection.authenticate()
        print('*** Authenticated')

        # Enable encryption
        print('*** Enabling encryption...')
        await connection.encrypt()
        print('*** Encryption on')

        # Look for an A2DP service
        avdtp_version = await find_a2dp_service(connection)
        if not avdtp_version:
            print(color('!!! no AVDTP service found'))
            return
        print(f'AVDTP version: {avdtp_version[0]}.{avdtp_version[1]}')

        # Create a client to interact with the remote device
        client = await AVDTP_Protocol.connect(connection, avdtp_version)

        # Discover all endpoints on the remote device
        endpoints = list(await client.discover_remote_endpoints())
        print(f'@@@ Found {len(endpoints)} endpoints')
        for endpoint in endpoints:
            print('@@@', endpoint)


# -----------------------------------------------------------------------------
bumble.logging.setup_basic_logging('DEBUG')
asyncio.run(main())



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Apple Media Service (AMS).
"""

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------
from __future__ import annotations

import asyncio
import dataclasses
import enum
import logging
from collections.abc import Iterable

from bumble import utils
from bumble.device import Peer
from bumble.gatt import (
    GATT_AMS_ENTITY_ATTRIBUTE_CHARACTERISTIC,
    GATT_AMS_ENTITY_UPDATE_CHARACTERISTIC,
    GATT_AMS_REMOTE_COMMAND_CHARACTERISTIC,
    GATT_AMS_SERVICE,
    Characteristic,
    TemplateService,
)
from bumble.gatt_client import CharacteristicProxy, ProfileServiceProxy, ServiceProxy

# -----------------------------------------------------------------------------
# Logging
# -----------------------------------------------------------------------------
logger = logging.getLogger(__name__)


# -----------------------------------------------------------------------------
# Protocol
# -----------------------------------------------------------------------------
class RemoteCommandId(utils.OpenIntEnum):
    PLAY = 0
    PAUSE = 1
    TOGGLE_PLAY_PAUSE = 2
    NEXT_TRACK = 3
    PREVIOUS_TRACK = 4
    VOLUME_UP = 5
    VOLUME_DOWN = 6
    ADVANCE_REPEAT_MODE = 7
    ADVANCE_SHUFFLE_MODE = 8
    SKIP_FORWARD = 9
    SKIP_BACKWARD = 10
    LIKE_TRACK = 11
    DISLIKE_TRACK = 12
    BOOKMARK_TRACK = 13


class EntityId(utils.OpenIntEnum):
    PLAYER = 0
    QUEUE = 1
    TRACK = 2


class ActionId(utils.OpenIntEnum):
    POSITIVE = 0
    NEGATIVE = 1


class EntityUpdateFlags(enum.IntFlag):
    TRUNCATED = 1


class PlayerAttributeId(utils.OpenIntEnum):
    NAME = 0
    PLAYBACK_INFO = 1
    VOLUME = 2


class QueueAttributeId(utils.OpenIntEnum):
    INDEX = 0
    COUNT = 1
    SHUFFLE_MODE = 2
    REPEAT_MODE = 3


class ShuffleMode(utils.OpenIntEnum):
    OFF = 0
    ONE = 1
    ALL = 2


class RepeatMode(utils.OpenIntEnum):
    OFF = 0
    ONE = 1
    ALL = 2


class TrackAttributeId(utils.OpenIntEnum):
    ARTIST = 0
    ALBUM = 1
    TITLE = 2
    DURATION = 3


class PlaybackState(utils.OpenIntEnum):
    PAUSED = 0
    PLAYING = 1
    REWINDING = 2
    FAST_FORWARDING = 3


@dataclasses.dataclass
class PlaybackInfo:
    playback_state: PlaybackState = PlaybackState.PAUSED
    playback_rate: float = 1.0
    elapsed_time: float = 0.0


# -----------------------------------------------------------------------------
# GATT Server-side
# -----------------------------------------------------------------------------
class Ams(TemplateService):
    UUID = GATT_AMS_SERVICE

    remote_command_characteristic: Characteristic
    entity_update_characteristic: Characteristic
    entity_attribute_characteristic: Characteristic

    def __init__(self) -> None:
        # TODO not the final implementation
        self.remote_command_characteristic = Characteristic(
            GATT_AMS_REMOTE_COMMAND_CHARACTERISTIC,
            Characteristic.Properties.NOTIFY
            | Characteristic.Properties.WRITE_WITHOUT_RESPONSE,
            Characteristic.Permissions.WRITEABLE,
        )

        # TODO not the final implementation
        self.entity_update_characteristic = Characteristic(
            GATT_AMS_ENTITY_UPDATE_CHARACTERISTIC,
            Characteristic.Properties.NOTIFY | Characteristic.Properties.WRITE,
            Characteristic.Permissions.WRITEABLE,
        )

        # TODO not the final implementation
        self.entity_attribute_characteristic = Characteristic(
            GATT_AMS_ENTITY_ATTRIBUTE_CHARACTERISTIC,
            Characteristic.Properties.READ
            | Characteristic.Properties.WRITE_WITHOUT_RESPONSE,
            Characteristic.Permissions.WRITEABLE | Characteristic.Permissions.READABLE,
        )

        super().__init__(
            [
                self.remote_command_characteristic,
                self.entity_update_characteristic,
                self.entity_attribute_characteristic,
            ]
        )


# -----------------------------------------------------------------------------
# GATT Client-side
# -----------------------------------------------------------------------------
class AmsProxy(ProfileServiceProxy):
    SERVICE_CLASS = Ams

    # NOTE: these don't use adapters, because the format for write and notifications
    # are different.
    remote_command: CharacteristicProxy[bytes]
    entity_update: CharacteristicProxy[bytes]
    entity_attribute: CharacteristicProxy[bytes]

    def __init__(self, service_proxy: ServiceProxy):
        self.remote_command = service_proxy.get_required_characteristic_by_uuid(
            GATT_AMS_REMOTE_COMMAND_CHARACTERISTIC
        )

        self.entity_update = service_proxy.get_required_characteristic_by_uuid(
            GATT_AMS_ENTITY_UPDATE_CHARACTERISTIC
        )

        self.entity_attribute = service_proxy.get_required_characteristic_by_uuid(
            GATT_AMS_ENTITY_ATTRIBUTE_CHARACTERISTIC
        )


class AmsClient(utils.EventEmitter):
    EVENT_SUPPORTED_COMMANDS = "supported_commands"
    EVENT_PLAYER_NAME = "player_name"
    EVENT_PLAYER_PLAYBACK_INFO = "player_playback_info"
    EVENT_PLAYER_VOLUME = "player_volume"
    EVENT_QUEUE_COUNT = "queue_count"
    EVENT_QUEUE_INDEX = "queue_index"
    EVENT_QUEUE_SHUFFLE_MODE = "queue_shuffle_mode"
    EVENT_QUEUE_REPEAT_MODE = "queue_repeat_mode"
    EVENT_TRACK_ARTIST = "track_artist"
    EVENT_TRACK_ALBUM = "track_album"
    EVENT_TRACK_TITLE = "track_title"
    EVENT_TRACK_DURATION = "track_duration"

    supported_commands: set[RemoteCommandId]
    player_name: str = ""
    player_playback_info: PlaybackInfo = PlaybackInfo(PlaybackState.PAUSED, 0.0, 0.0)
    player_volume: float = 1.0
    queue_count: int = 0
    queue_index: int = 0
    queue_shuffle_mode: ShuffleMode = ShuffleMode.OFF
    queue_repeat_mode: RepeatMode = RepeatMode.OFF
    track_artist: str = ""
    track_album: str = ""
    track_title: str = ""
    track_duration: float = 0.0

    def __init__(self, ams_proxy: AmsProxy) -> None:
        super().__init__()
        self._ams_proxy = ams_proxy
        self._started = False
        self._read_attribute_semaphore = asyncio.Semaphore()
        self.supported_commands = set()

    @classmethod
    async def for_peer(cls, peer: Peer) -> AmsClient | None:
        ams_proxy = await peer.discover_service_and_create_proxy(AmsProxy)
        if ams_proxy is None:
            return None
        return cls(ams_proxy)

    async def start(self) -> None:
        logger.debug("subscribing to remote command characteristic")
        await self._ams_proxy.remote_command.subscribe(
            self._on_remote_command_notification
        )

        logger.debug("subscribing to entity update characteristic")
        await self._ams_proxy.entity_update.subscribe(
            lambda data: utils.AsyncRunner.spawn(
                self._on_entity_update_notification(data)
            )
        )

        self._started = True

    async def stop(self) -> None:
        await self._ams_proxy.remote_command.unsubscribe(
            self._on_remote_command_notification
        )
        await self._ams_proxy.entity_update.unsubscribe(
            self._on_entity_update_notification
        )
        self._started = False

    async def observe(
        self,
        entity: EntityId,
        attributes: Iterable[PlayerAttributeId | QueueAttributeId | TrackAttributeId],
    ) -> None:
        await self._ams_proxy.entity_update.write_value(
            bytes([entity] + list(attributes)), with_response=True
        )

    async def command(self, command: RemoteCommandId) -> None:
        await self._ams_proxy.remote_command.write_value(
            bytes([command]), with_response=True
        )

    async def play(self) -> None:
        await self.command(RemoteCommandId.PLAY)

    async def pause(self) -> None:
        await self.command(RemoteCommandId.PAUSE)

    async def toggle_play_pause(self) -> None:
        await self.command(RemoteCommandId.TOGGLE_PLAY_PAUSE)

    async def next_track(self) -> None:
        await self.command(RemoteCommandId.NEXT_TRACK)

    async def previous_track(self) -> None:
        await self.command(RemoteCommandId.PREVIOUS_TRACK)

    async def volume_up(self) -> None:
        await self.command(RemoteCommandId.VOLUME_UP)

    async def volume_down(self) -> None:
        await self.command(RemoteCommandId.VOLUME_DOWN)

    async def advance_repeat_mode(self) -> None:
        await self.command(RemoteCommandId.ADVANCE_REPEAT_MODE)

    async def advance_shuffle_mode(self) -> None:
        await self.command(RemoteCommandId.ADVANCE_SHUFFLE_MODE)

    async def skip_forward(self) -> None:
        await self.command(RemoteCommandId.SKIP_FORWARD)

    async def skip_backward(self) -> None:
        await self.command(RemoteCommandId.SKIP_BACKWARD)

    async def like_track(self) -> None:
        await self.command(RemoteCommandId.LIKE_TRACK)

    async def dislike_track(self) -> None:
        await self.command(RemoteCommandId.DISLIKE_TRACK)

    async def bookmark_track(self) -> None:
        await self.command(RemoteCommandId.BOOKMARK_TRACK)

    def _on_remote_command_notification(self, data: bytes) -> None:
        supported_commands = [RemoteCommandId(command) for command in data]
        logger.debug(
            f"supported commands: {[command.name for command in supported_commands]}"
        )
        for command in supported_commands:
            self.supported_commands.add(command)

        self.emit(self.EVENT_SUPPORTED_COMMANDS)

    async def _on_entity_update_notification(self, data: bytes) -> None:
        entity = EntityId(data[0])
        flags = EntityUpdateFlags(data[2])
        value = data[3:]

        if flags & EntityUpdateFlags.TRUNCATED:
            logger.debug("truncated attribute, fetching full value")

            # Write the entity and attribute we're interested in
            # (protected by a semaphore, so that we only read one attribute at a time)
            async with self._read_attribute_semaphore:
                await self._ams_proxy.entity_attribute.write_value(
                    data[:2], with_response=True
                )
                value = await self._ams_proxy.entity_attribute.read_value()

        if entity == EntityId.PLAYER:
            player_attribute = PlayerAttributeId(data[1])
            if player_attribute == PlayerAttributeId.NAME:
                self.player_name = value.decode()
                self.emit(self.EVENT_PLAYER_NAME)
            elif player_attribute == PlayerAttributeId.PLAYBACK_INFO:
                playback_state_str, playback_rate_str, elapsed_time_str = (
                    value.decode().split(",")
                )
                self.player_playback_info = PlaybackInfo(
                    PlaybackState(int(playback_state_str)),
                    float(playback_rate_str),
                    float(elapsed_time_str),
                )
                self.emit(self.EVENT_PLAYER_PLAYBACK_INFO)
            elif player_attribute == PlayerAttributeId.VOLUME:
                self.player_volume = float(value.decode())
                self.emit(self.EVENT_PLAYER_VOLUME)
            else:
                logger.warning(f"received unknown player attribute {player_attribute}")

        elif entity == EntityId.QUEUE:
            queue_attribute = QueueAttributeId(data[1])
            if queue_attribute == QueueAttributeId.COUNT:
                self.queue_count = int(value)
                self.emit(self.EVENT_QUEUE_COUNT)
            elif queue_attribute == QueueAttributeId.INDEX:
                self.queue_index = int(value)
                self.emit(self.EVENT_QUEUE_INDEX)
            elif queue_attribute == QueueAttributeId.REPEAT_MODE:
                self.queue_repeat_mode = RepeatMode(int(value))
                self.emit(self.EVENT_QUEUE_REPEAT_MODE)
            elif queue_attribute == QueueAttributeId.SHUFFLE_MODE:
                self.queue_shuffle_mode = ShuffleMode(int(value))
                self.emit(self.EVENT_QUEUE_SHUFFLE_MODE)
            else:
                logger.warning(f"received unknown queue attribute {queue_attribute}")

        elif entity == EntityId.TRACK:
            track_attribute = TrackAttributeId(data[1])
            if track_attribute == TrackAttributeId.ARTIST:
                self.track_artist = value.decode()
                self.emit(self.EVENT_TRACK_ARTIST)
            elif track_attribute == TrackAttributeId.ALBUM:
                self.track_album = value.decode()
                self.emit(self.EVENT_TRACK_ALBUM)
            elif track_attribute == TrackAttributeId.TITLE:
                self.track_title = value.decode()
                self.emit(self.EVENT_TRACK_TITLE)
            elif track_attribute == TrackAttributeId.DURATION:
                self.track_duration = float(value.decode())
                self.emit(self.EVENT_TRACK_DURATION)
            else:
                logger.warning(f"received unknown track attribute {track_attribute}")

        else:
            logger.warning(f"received unknown attribute ID {data[1]}")
